/*
 * DoubleCheck, a UCI chess playing engine. Copyright (C) 2011 Lucas Braesch.
 *
 * DoubleCheck is free software: you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * DoubleCheck is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not,
 * see <http://www.gnu.org/licenses/>.
*/
#include "bitboard.h"
bool BitboardInitialized = false;

static void init_masks();
static void init_attacks();
static void init_rays();

void init_bitboard()
{
	if (BitboardInitialized) return;

	init_magics();	
	init_masks();
	init_attacks();
	init_rays();
	
	BitboardInitialized = true;
}

/* General purpose bitboards */

uint64_t Bit[NB_SQUARE], Rank[NB_RANK_FILE], File[NB_RANK_FILE];
uint64_t PInitialRank[NB_COLOR], PPromotionRank[NB_COLOR];
uint64_t AdjacentFiles[NB_RANK_FILE], ThisAndAdjacentFiles[NB_RANK_FILE];
uint64_t InFront[NB_COLOR][NB_RANK_FILE], SquaresInFront[NB_COLOR][NB_SQUARE];
uint64_t PassedPawnMask[NB_COLOR][NB_SQUARE], PawnSpanMask[NB_COLOR][NB_SQUARE];
uint64_t KingDistance[NB_SQUARE][NB_SQUARE];

static void init_masks()
{
	for (unsigned i = A1; i <= H8; i++)
		Bit[i] = 1ULL << i;

	for (unsigned i = 0; i < NB_RANK_FILE; i++) {
		Rank[i] = 0xFFULL << (8 * i);
		File[i] = 0x0101010101010101ULL << i;
	}

	for (unsigned f = FileA; f <= FileH; f++) {
		AdjacentFiles[f] = 0ULL;
		if (f > FileA) AdjacentFiles[f] |= File[f-1];
		if (f < FileH) AdjacentFiles[f] |= File[f+1];
		ThisAndAdjacentFiles[f] = File[f] | AdjacentFiles[f];
	}

	InFront[White][Rank8] = InFront[Black][Rank1] = 0ULL;
	for (int rw = Rank7, rb = Rank2; rw >= Rank1; rw--, rb++) {
		InFront[White][rw] = InFront[White][rw+1] | Rank[rw+1];
		InFront[Black][rb] = InFront[Black][rb-1] | Rank[rb-1];
	}

	PInitialRank[White] = Rank[Rank2];
	PInitialRank[Black] = Rank[Rank7];
	PPromotionRank[White] = Rank[Rank8];
	PPromotionRank[Black] = Rank[Rank1];

	for (unsigned color = White; color <= Black; color++)
		for (unsigned sq = A1; sq <= H8; sq++) {
			PassedPawnMask[color][sq] = InFront[color][rank(sq)] & ThisAndAdjacentFiles[file(sq)];
			PawnSpanMask[color][sq] = InFront[color][rank(sq)] & AdjacentFiles[file(sq)];
			SquaresInFront[color][sq] = PassedPawnMask[color][sq] & File[file(sq)];
		}

	for (unsigned s1 = A1; s1 <= H8; s1++)
		for (unsigned s2 = A1; s2 <= H8; s2++)
			KingDistance[s1][s2] = max(abs(file(s1)-file(s2)), abs(rank(s1)-rank(s2)));
}

/* Piece attacks and pseudo attacks bitboards */

uint64_t KAttacks[NB_SQUARE], NAttacks[NB_SQUARE], PAttacks[NB_COLOR][NB_SQUARE];
uint64_t BPseudoAttacks[NB_SQUARE], RPseudoAttacks[NB_SQUARE];
uint64_t PawnsAttacking[NB_COLOR][NB_SQUARE];
uint64_t Shield[NB_COLOR][NB_SQUARE];

static uint64_t calc_KAttacks(unsigned sq)
{
	assert(square_ok(sq));
	const uint64_t b = Bit[sq];
	return	((b & ~FileA_bb) << 7)
		|	((b & ~FileA_bb) >> 1)
		|	((b & ~FileA_bb) >> 9)
		|	((b & ~FileH_bb) << 9)
		|	((b & ~FileH_bb) << 1)
		|	((b & ~FileH_bb) >> 7)
		|	(b << 8)
		|	(b >> 8);
}

static uint64_t calc_NAttacks(unsigned sq)
{
	assert(square_ok(sq));
	const uint64_t b = Bit[sq];
	return	((b & ~FileA_bb) << 15)
		|	((b & ~FileH_bb) << 17)
		|	((b & ~FileA_bb) >> 17)
		|	((b & ~FileH_bb) >> 15)
		|	((b & ~(FileA_bb | File[FileB])) << 6)
		|	((b & ~(FileA_bb | File[FileB])) >> 10)
		|	((b & ~(File[FileG] | FileH_bb)) << 10)
		|	((b & ~(File[FileG] | FileH_bb)) >> 6);
}

static uint64_t calc_PAttacks(unsigned color, unsigned sq)
{
	assert(color_ok(color) && square_ok(sq));
	const uint64_t b = Bit[sq];
	return	shift_bit(b & ~FileA_bb, color ? -9 : +7)
		|	shift_bit(b & ~FileH_bb, color ? -7 : +9);
}

static void init_attacks()
{
	for (unsigned sq = A1; sq <= H8; sq++) {
		KAttacks[sq] = calc_KAttacks(sq);
		NAttacks[sq] = calc_NAttacks(sq);
		BPseudoAttacks[sq] = bishop_attack(sq, 0);
		RPseudoAttacks[sq] = rook_attack(sq, 0);
		
		for (unsigned color = White; color <= Black; color++)
			PAttacks[color][sq] = calc_PAttacks(color, sq);		
	}
	
	for (unsigned sq = A1; sq <= H8; sq++) {
		for (unsigned color = White; color <= Black; color++) {
			Shield[color][sq] = KAttacks[sq] & PassedPawnMask[color][sq];
			
			PawnsAttacking[color][sq] = 0;
			uint64_t ring = KAttacks[sq];
			while (ring) {
				const unsigned target = next_bit(&ring);
				PawnsAttacking[color][sq] |= PAttacks[color][target];
			}
		}		
	}
}

/* Ray bitboards */

uint64_t Between[NB_SQUARE][NB_SQUARE], Direction[NB_SQUARE][NB_SQUARE];

static void init_rays()
{
	const int dir[8][2] = {{-1,-1},{-1,0},{-1,1},{0,-1},{0,1},{1,-1},{1,0},{1,1}};
	memset(Between, 0, sizeof(Between));
	memset(Direction, 0, sizeof(Direction));

	for (unsigned sq = A1; sq <= H8; sq++) {		
		unsigned r = rank(sq), f = file(sq);

		for (unsigned i = 0; i < 8; i++) {
			int dr = dir[i][0], df = dir[i][1];
			int _r, _f, _sq;
			uint64_t mask;

			for (_r=r+dr,_f=f+df,mask=0ULL; rank_file_ok(_r,_f); _r+=dr,_f+=df) {
				_sq = square(_r,_f);
				mask |= 1ULL << _sq;
				Between[sq][_sq] = mask;
			}

			uint64_t direction = mask;

			while (mask) {
				_sq = next_bit(&mask);
				Direction[sq][_sq] = direction;
			}
		}
	}	
}

static const unsigned BitTable[NB_SQUARE] = {
	0, 1, 2, 7, 3, 13, 8, 19, 4, 25, 14, 28, 9, 34, 20, 40, 5, 17, 26, 38, 15,
	46, 29, 48, 10, 31, 35, 54, 21, 50, 41, 57, 63, 6, 12, 18, 24, 27, 33, 39,
	16, 37, 45, 47, 30, 53, 49, 56, 62, 11, 23, 32, 36, 44, 52, 55, 61, 22, 43,
	51, 60, 42, 59, 58
};

unsigned first_bit(uint64_t b)
/* From Stockfish */
{
	return BitTable[((b & -b) * 0x218a392cd3d5dbfULL) >> 58];
}

unsigned next_bit(uint64_t *b)
/* From Stockfish */
{
	uint64_t _b = *b;
	*b &= *b - 1;
	return BitTable[((_b & -_b) * 0x218a392cd3d5dbfULL) >> 58];
}

unsigned count_bit(uint64_t b)
/* General purpose bit counting, using a simple loop */
{
	unsigned count = 0;
	while (b) {
		b &= b - 1;
		count++;
	}
	return count;
}

unsigned count_bit_max15(uint64_t b)
/* From Stockfish: count bits up to 15, without a loop */
{
	assert(count_bit(b) <= 15);
	b -= (b>>1) & 0x5555555555555555ULL;
	b = ((b>>2) & 0x3333333333333333ULL) + (b & 0x3333333333333333ULL);
	b *= 0x1111111111111111ULL;
	return b >> 60;
}

void print_bitboard(uint64_t b)
{
	assert(BitboardInitialized);
	
	for (int r = Rank8; r >= Rank1; r--) {
		for (int f = FileA; f <= FileH; f++) {
			unsigned sq = square(r, f);
			char c = test_bit(b, sq) ? 'X' : '.';
			printf(" %c", c);
		}
		printf("\n");
	}
}