/*
 * DoubleCheck, a UCI chess playing engine. Copyright (C) 2011 Lucas Braesch.
 *
 * DoubleCheck is free software: you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * DoubleCheck is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not,
 * see <http://www.gnu.org/licenses/>.
*/
#pragma once
#include "defs.h"

void init_bitboard();
extern bool BitboardInitialized;

/* Bitmask for each square, rank and file */
extern uint64_t Bit[NB_SQUARE], Rank[NB_RANK_FILE], File[NB_RANK_FILE];

/* Most commonly used file bitboards. Equivalent to File[FileA] and File[FileH], but faster */
static const uint64_t FileA_bb = 0x0101010101010101ULL << FileA;
static const uint64_t FileH_bb = 0x0101010101010101ULL << FileH;

/* Between[s1][s2] is the segment ]s1,s2] when then angle (s1,s2) is a multiple of 45 degrees, 0
 * otherwise. Direction[s1][s2] is the the half-line from s1 to s2 going all the way to the edge of
 * the board. */
extern uint64_t Between[NB_SQUARE][NB_SQUARE], Direction[NB_SQUARE][NB_SQUARE];

/* KAttacks[sq], NAttacks[sq], PAttacks[sq] are the set of squares attacked by a King, Knight, Pawn
 * of color, on sq. */
extern uint64_t KAttacks[NB_SQUARE], NAttacks[NB_SQUARE], PAttacks[NB_COLOR][NB_SQUARE];

/* BPseudoAttacks[NB_SQUARE], RPseudoAttacks[NB_SQUARE] are the set of squares attacked by a Bishop,
 * Rook on sq, given an empty board. */
extern uint64_t BPseudoAttacks[NB_SQUARE], RPseudoAttacks[NB_SQUARE];

/* PInitialRank[color] is the initial Pawn rank for color.
 * PPromotionRank[color] is the rank on which Pawns of color get promoted. */
extern uint64_t PInitialRank[NB_COLOR], PPromotionRank[NB_COLOR];

/* AdjacentFiles[f] is the set of squares represented by the two files adjacent to f. Obviously this
 * is only one file when f = FileA or f = FileH.
 * ThisAdjacentFiles[f] is the same with File[f] added (so 3 files, and 2 when on the edge). */
extern uint64_t AdjacentFiles[NB_RANK_FILE], ThisAndAdjacentFiles[NB_RANK_FILE];

/* InFront[r] represent all the ranks strictly above Rank[r].
 * SquaresInFront[color][sq] represent the squares on the same file, that are strictly "in front" of
 * sq from the perspective of color. IOW it's the promotion path of a pawn of color on sq. */
extern uint64_t InFront[NB_COLOR][NB_RANK_FILE], SquaresInFront[NB_COLOR][NB_SQUARE];

/* PassedPawnMask[color][sq] = InFront[color][rank(sq)] & ThisAndAdjacentFiles[file(sq)]
 * PawnSpanMask[color][sq] = InFront[color][rank(sq)] & AdjacentFiles[file(sq)] */
extern uint64_t PassedPawnMask[NB_COLOR][NB_SQUARE], PawnSpanMask[NB_COLOR][NB_SQUARE];

/* Each color's half board: Rank[Rank1..4] for White and Rank[Rank5..8] for Black */
static const uint64_t HalfBoard[NB_COLOR] = {0xFFFFFFFFULL, 0xFFFFFFFFULL<<32};

/* Enemy Pawns attacking the ring around a King of color on sq */
extern uint64_t PawnsAttacking[NB_COLOR][NB_SQUARE];

/* 3 squares in front of a King of color on sq. Useful for King safety code */
extern uint64_t Shield[NB_COLOR][NB_SQUARE];

/* Accessors to precalculated bitboard arrays. Make calling code easier, and allows for assert()
 * runtime checking in debug mode. */

static inline uint64_t file_bb(unsigned sq)
{
	assert(BitboardInitialized && square_ok(sq));
	return File[file(sq)];
}

static inline uint64_t rank_bb(unsigned sq)
{
	assert(BitboardInitialized && square_ok(sq));
	return Rank[rank(sq)];
}

static inline uint64_t adjacent_files_bb(unsigned sq)
{
	assert(BitboardInitialized && square_ok(sq));
	return AdjacentFiles[file(sq)];
}

static inline uint64_t this_and_adjacent_files_bb(unsigned sq)
{
	assert(BitboardInitialized && square_ok(sq));
	return ThisAndAdjacentFiles[file(sq)];
}

static inline uint64_t in_front_bb(unsigned color, unsigned sq)
{
	assert(BitboardInitialized && color_ok(color));
	return InFront[color][rank(sq)];
}

static inline unsigned pawn_push(unsigned color, unsigned sq)
/* Destination square for a Pawn of color to be pushed */
{
	assert(color_ok(color) && rank(sq) >= Rank2 && rank(sq) <= Rank7);
	return color ? sq-8 : sq+8;
}

/* KingDistance[s1][s2] is the King distance from s1 to s2, assuming empty board */
uint64_t KingDistance[NB_SQUARE][NB_SQUARE];

static inline unsigned king_dist(unsigned s1, unsigned s2)
{
	assert(BitboardInitialized && square_ok(s1) && square_ok(s2));
	return KingDistance[s1][s2];
}

/* Displays a bitboard on stdout: 'X' when a square is occupied and '.' otherwise */
void print_bitboard(uint64_t b);

/* Functions to set, clear, test, and shift bits. Of course, it can be done the hacker-way (directly
 * with & ~ <<), but these functions perform runtime assert() in debug mode, and make the calling
 * code readable. */

static inline void set_bit(uint64_t *b, unsigned i)
{
	assert(BitboardInitialized && square_ok(i));
	*b |= Bit[i];
}

static inline void clear_bit(uint64_t *b, unsigned i)
{
	assert(BitboardInitialized && square_ok(i));
	*b &= ~Bit[i];
}

static inline uint64_t test_bit(uint64_t b, unsigned i)
{
	assert(BitboardInitialized && square_ok(i));
	return b & Bit[i];
}

static inline uint64_t shift_bit(uint64_t b, int i)
/* shift_bit() extends << in allowing negative shifts. This is useful to keep some pawn related code
 * generic and simple. */
{
	assert(abs(i) < 64);
	return i > 0 ? b << i : b >> -i;
}

/* true if b has no more than one bit set */
static inline bool single_bit(uint64_t b)
{ return !(b & (b - 1)); }

/* Return the lowest order bit set in b */
unsigned first_bit(uint64_t b);

/* Return the lowest order bit set in b, and removes this bit from b: use this of bit loops */
unsigned next_bit(uint64_t *b);

/* Counts the number of bits set in b, using a loop. Performs best on sparsly populated bitboards */
unsigned count_bit(uint64_t b);

/* Counts the number of bits set in b, up to a maximum of 15. Faster than count_bit for not so
 * sparsly populated bitboards */
unsigned count_bit_max15(uint64_t b);

/* Precalculate magic bitboards */
void init_magics();
extern bool MagicInitialized;

/* Squares attacked by a bishop/rook for a given board occupancy */
extern uint64_t bishop_attack(unsigned sq, uint64_t occ);
extern uint64_t rook_attack(unsigned sq, uint64_t occ);

static inline uint64_t piece_attack(unsigned piece, unsigned sq, uint64_t occ)
/* Generic attack function for pieces (not pawns). Typically, this is used in a block that loops on
 * piece, so inling this allows some optimizations in the calling code, thanks to loop unrolling */
{
	assert(BitboardInitialized && MagicInitialized);
	assert(Knight <= piece && piece <= King && square_ok(sq));

	if (piece == Knight)
		return NAttacks[sq];
	else if (piece == Bishop)
		return bishop_attack(sq, occ);
	else if (piece == Rook)
		return rook_attack(sq, occ);
	else if (piece == Queen)
		return bishop_attack(sq, occ) | rook_attack(sq, occ);
	else	// if (piece == King)
		return KAttacks[sq];
}