/*
 * DoubleCheck, a UCI chess playing engine. Copyright (C) 2011 Lucas Braesch.
 * 
 * DoubleCheck is free software: you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * DoubleCheck is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program. If not,
 * see <http://www.gnu.org/licenses/>.
*/
#include "board.h"
#include "move.h"
#include "zobrist.h"
#include "psq.h"

const char *PieceLabel[NB_COLOR] = { "PNBRQK", "pnbrqk" };
const char *StartFen = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1";

static void calc_attacks(const Board *B);
static uint64_t calc_checkers(const Board *B, unsigned kcolor);
static uint64_t hidden_checkers(const Board *B, unsigned find_pins, unsigned color);
static int full_see(const Board *B, unsigned fsq, unsigned tsq);

void reset_board(Board *B)
{
	assert(BitboardInitialized && ZobristInitialized);
	memset(B, 0, sizeof(Board));

	for (unsigned sq = A1; sq <= H8; B->piece_on[sq++] = NoPiece);
	B->st = B->game_stack;
	B->st->epsq = NoSquare;
	B->st->last_move = NoMove;
	
	B->initialized = true;
}

void load_fen(Board *B, const char *fen)
{
	assert(fen);
	unsigned r, f;
	char c = *fen++;

	// load the board
	reset_board(B);
	for (r = Rank8, f = FileA; c != ' ' && r >= Rank1; c = *fen++) {
		// empty squares
		if ('1' <= c && c <= '8') {
			f += c - '0';
			assert(f <= NB_RANK_FILE);
		}
		// EOL separator
		else if (c == '/') {
			assert(f == NB_RANK_FILE);
			r--;
			f = FileA;
		}
		// piece
		else {
			unsigned sq = square(r, f++), piece;
			for (piece = Pawn; piece <= King; piece++) {
				if (c == PieceLabel[isupper(c) ? White : Black][piece])
					break;
			}
			assert(piece_ok(piece));
			unsigned color = isupper(c) ? White : Black;
			set_square(B, color, piece, sq, true);
			if (piece == King)
				B->king_pos[color] = sq;
		}
	}

	// turn of play
	c = *fen++;
	assert(c == 'w' || c == 'b');
	B->turn = c == 'w' ? White : Black;
	B->st->key ^= B->turn ? zobrist_turn : 0ULL;
	const unsigned us = B->turn, them = opp_color(us);

	// Castling rights
	if (*fen++ != ' ') assert(0);
	c = *fen++;
	B->st->crights = 0;

	if (c == '-')	// no castling rights
		c = *fen++;
	else {			// set castling rights (in reading order KQkq)
		if (c == 'K') {
			B->st->crights ^= make_crights(White, OO);
			c = *fen++;
		}
		if (c == 'Q') {
			B->st->crights ^= make_crights(White, OOO);
			c = *fen++;
		}
		if (c == 'k') {
			B->st->crights ^= make_crights(Black, OO);
			c = *fen++;
		}
		if (c == 'q') {
			B->st->crights ^= make_crights(Black, OOO);
			c = *fen++;
		}
	}

	// en passant square
	assert(c == ' ');
	c = *fen++;
	if (c == '-') {
		// no en passant square
		c = *fen++;
		B->st->epsq = NoSquare;
	} else {
		unsigned r, f;
		assert('a' <= c && c <= 'h');
		f = c - 'a';
		c = *fen++;
		assert(isdigit(c));
		r = c - '1';
		B->st->epsq = square (r, f);
	}

	B->st->pinned = hidden_checkers(B, 1, us);
	B->st->dcheckers = hidden_checkers(B, 0, us);
	calc_attacks(B);
	B->st->attacked = B->st->attacks[them][NoPiece];
	B->st->checkers = test_bit(B->st->attacked, B->king_pos[us]) ? calc_checkers(B, us) : 0ULL;

	assert(calc_key(B) == B->st->key);
	assert(calc_kpkey(B) == B->st->kpkey);
	assert(psq_ok(B));
}

void write_fen(const Board *B, char *fen)
{
	assert(B->initialized && fen);

	// write the board
	for (int r = Rank8; r >= Rank1; r--) {
		unsigned empty_cnt = 0;
		for (int f = FileA; f <= FileH; f++) {
			unsigned sq = square(r, f);
			if (B->piece_on[sq] == NoPiece)
				empty_cnt++;
			else {
				if (empty_cnt) {
					*fen++ = empty_cnt + '0';
					empty_cnt = 0;
				}
				*fen++ = PieceLabel[color_on(B, sq)][B->piece_on[sq]];
			}
		}
		if (empty_cnt) *fen++ = empty_cnt + '0';
		*fen++ = r == Rank1 ? ' ' : '/';
	}

	// turn of play
	*fen++ = B->turn ? 'b' : 'w';
	*fen++ = ' ';

	// castling rights
	unsigned crights = B->st->crights;
	if (crights) {
		if (crights & make_crights(White, OO))	*fen++ = 'K';
		if (crights & make_crights(White, OOO))	*fen++ = 'Q';
		if (crights & make_crights(Black, OO))	*fen++ = 'k';
		if (crights & make_crights(Black, OOO))	*fen++ = 'q';
	} else
		*fen++ = '-';
	*fen++ = ' ';

	// en passant square
	unsigned epsq = B->st->epsq;
	if (epsq != NoSquare) {
		*fen++ = file(epsq) + 'a';
		*fen++ = rank(epsq) + '1';
	} else
		*fen++ = '-';

	*fen++ = '\0';
}

void print_board(const Board *B)
{
	assert(B->initialized);
	
	for (int r = Rank8; r >= Rank1; r--) {
		for (int f = FileA; f <= FileH; f++) {
			unsigned sq = square(r, f),
			    color = color_on(B, sq);
			char c = color != NoColor
			         ? PieceLabel[color][B->piece_on[sq]]
			         : (sq == B->st->epsq ? '*' : '.');
			printf(" %c", c);
		}
		printf("\n");
	}
	printf("\n");
}

void play(Board *B, move_t m)
{
	assert(B->initialized);
	
	++B->st;
	memcpy(B->st, B->st-1, sizeof(game_info));
	B->st->last_move = m;
	B->st->rule50++;

	const unsigned us = B->turn, them = opp_color(us);
	const unsigned piece = B->piece_on[m.fsq], capture = B->piece_on[m.tsq];

	if (move_equal(m, NoMove)) {
		assert(!board_is_check(B));
		B->st->epsq = NoSquare;
	} else {
		// normal capture: remove captured piece
		if (piece_ok(capture)) {
			B->st->rule50 = 0;
			clear_square(B, them, capture, m.tsq, true);
		}

		// move our piece
		clear_square(B, us, piece, m.fsq, true);
		set_square(B, us, m.promotion == NoPiece ? piece : m.promotion, m.tsq, true);

		if (piece == Pawn) {
			B->st->rule50 = 0;
			int inc_pp = us ? -8 : 8;
			// set the epsq if double push
			B->st->epsq = (m.tsq - m.fsq == 2 * inc_pp)
			              ? m.fsq + inc_pp
			              : NoSquare;
			// capture en passant
			if (m.ep)
				clear_square(B, them, Pawn, m.tsq - inc_pp, true);
		} else {
			B->st->epsq = NoSquare;

			if (piece == Rook) {
				// a rook move can alter castling rights
				if (m.fsq == (us ? H8 : H1))
					B->st->crights &= ~make_crights(us, OO);
				else if (m.fsq == (us ? A8 : A1))
					B->st->crights &= ~make_crights(us, OOO);
			} else if (piece == King) {
				// update king_pos and clear crights
				B->king_pos[us] = m.tsq;
				B->st->crights &= ~make_crights(us, OO | OOO);
				// move the rook (jump over the king)
				if (m.tsq == m.fsq+2) {			// OO
					clear_square(B, us, Rook, us ? H8 : H1, true);
					set_square(B, us, Rook, us ? F8 : F1, true);
				} else if (m.tsq == m.fsq-2) {	// OOO
					clear_square(B, us, Rook, us ? A8 : A1, true);
					set_square(B, us, Rook, us ? D8 : D1, true);
				}
			}
		}

		if (capture == Rook) {
			// Rook captures can alter opponent's castling rights
			if (m.tsq == (us ? A1 : A8))
				B->st->crights &= ~make_crights(them, OOO);
			else if (m.tsq == (us ? H1 : H8))
				B->st->crights &= ~make_crights(them, OO);
		}
	}

	B->turn = them;
	B->st->key ^= zobrist_turn;

	B->st->piece = piece;
	B->st->capture = capture;
	
	B->st->pinned = hidden_checkers(B, 1, them);
	B->st->dcheckers = hidden_checkers(B, 0, them);
	calc_attacks(B);
	B->st->attacked = B->st->attacks[us][NoPiece];
	B->st->checkers = test_bit(B->st->attacked, B->king_pos[them]) ? calc_checkers(B, them) : 0ULL;

	assert(calc_key(B) == B->st->key);
	assert(calc_kpkey(B) == B->st->kpkey);
	assert(psq_ok(B));
}

void undo(Board *B)
{
	assert(B->initialized);
	const unsigned us = opp_color(B->turn), them = B->turn;
	const move_t m = B->st->last_move;
	const unsigned piece = B->st->piece, capture = B->st->capture;

	if (move_equal(m, NoMove))
		assert(!board_is_check(B));
	else {
		// move our piece back
		clear_square(B, us, m.promotion == NoPiece ? piece : m.promotion, m.tsq, false);
		set_square(B, us, piece, m.fsq, false);

		// restore the captured piece (if any)
		if (piece_ok(capture))
			set_square(B, them, capture, m.tsq, false);

		if (piece == King) {
			// update king_pos
			B->king_pos[us] = m.fsq;
			// undo rook jump (castling)
			if (m.tsq == m.fsq+2) {			// OO
				clear_square(B, us, Rook, us ? F8 : F1, false);
				set_square(B, us, Rook, us ? H8 : H1, false);
			} else if (m.tsq == m.fsq-2) {	// OOO
				clear_square(B, us, Rook, us ? D8 : D1, false);
				set_square(B, us, Rook, us ? A8 : A1, false);
			}
		} else if (m.ep)	// restore the en passant captured pawn
			set_square(B, them, Pawn, m.tsq + (us ? 8 : -8), false);
	}

	B->turn = us;
	--B->st;
}

uint64_t hidden_checkers(const Board *B, unsigned find_pins, unsigned color)
{
	assert(B->initialized && color_ok(color) && (find_pins == 0 || find_pins == 1));
	const unsigned aside = color ^ find_pins, kside = opp_color(aside);
	uint64_t result = 0ULL, pinners;

	// Pinned pieces protect our king, dicovery checks attack the enemy king.
	const unsigned ksq = B->king_pos[kside];

	// Pinners are only sliders with X-ray attacks to ksq
	pinners = (get_RQ(B, aside) & RPseudoAttacks[ksq]) | (get_BQ(B, aside) & BPseudoAttacks[ksq]);

	while (pinners) {
		unsigned sq = next_bit(&pinners);
		uint64_t b = Between[ksq][sq] & ~Bit[sq] & B->st->occ;
		// NB: if b == 0 then we're in check

		if (single_bit(b) && (b & B->all[color]))
			result |= b;
	}
	return result;
}

static void calc_attacks(const Board *B)
{
	assert(B->initialized);
	
	for (unsigned color = White; color <= Black; color++) {
		// King
		B->st->attacks[color][NoPiece] = B->st->attacks[color][King] = KAttacks[B->king_pos[color]];

		// Knight
		uint64_t fss = B->b[color][Knight];
		while (fss) {
			const unsigned fsq = next_bit(&fss);
			B->st->attacks[color][NoPiece] |= B->st->attacks[color][Knight] = NAttacks[fsq];
		}
		
		// piece attacks
		for (unsigned piece = Bishop; piece <= Queen; piece++) {
			uint64_t fss = B->b[color][piece];
			while (fss) {
				const unsigned fsq = next_bit(&fss);
				B->st->attacks[color][NoPiece] |=
					B->st->attacks[color][piece] = piece_attack(piece, fsq, B->st->occ);
			}
		}
		
		// pawn attacks
		B->st->attacks[color][NoPiece] |= B->st->attacks[color][Pawn]
			= shift_bit((B->b[color][Pawn] & ~FileA_bb), color ? -9 : 7);
		B->st->attacks[color][NoPiece] |= B->st->attacks[color][Pawn]
			|= shift_bit((B->b[color][Pawn] & ~FileH_bb), color ? -7 : 9);
	}
}

static uint64_t calc_checkers(const Board *B, unsigned kcolor)
{
	assert(B->initialized && color_ok(kcolor));
	const unsigned
		kpos = B->king_pos[kcolor],
		them = opp_color(kcolor);
	const uint64_t
		RQ = get_RQ(B, them) & RPseudoAttacks[kpos],
		BQ = get_BQ(B, them) & BPseudoAttacks[kpos];

	return (RQ & rook_attack(kpos, B->st->occ))
		| (BQ & bishop_attack(kpos, B->st->occ))
		| (B->b[them][Knight] & NAttacks[kpos])
		| (B->b[them][Pawn] & PAttacks[kcolor][kpos]);
}

static uint64_t calc_attackers(const Board *B, unsigned sq, uint64_t occ)
{
	assert(B->initialized && square_ok(sq));

	return (get_all_RQ(B) & rook_attack(sq, occ))
		| (get_all_BQ(B) & bishop_attack(sq, occ))
		| (NAttacks[sq] & (B->b[White][Knight] | B->b[Black][Knight]))
		| (KAttacks[sq] & (B->b[White][King] | B->b[Black][King]))
		| (PAttacks[White][sq] & B->b[Black][Pawn])
		| (PAttacks[Black][sq] & B->b[White][Pawn]);
}

static int full_see(const Board *B, unsigned fsq, unsigned tsq)
// SEE from StockFish: adapted and improved to handle promotions
{
	assert(B->initialized && square_ok(fsq) && square_ok(tsq));

	unsigned stm = color_on(B, fsq);	// side to move
	uint64_t attackers, stm_attackers;
	int swap_list[32], sl_idx = 1;
	uint64_t occ = B->st->occ;

	unsigned piece = B->piece_on[fsq], capture;

	const bool is_prank = test_bit(PPromotionRank[stm], tsq);
	static const unsigned see_val[NB_PIECE+1] = {MP, MN, MB, MR, MQ, 200*MP, 0};

	// Determine captured piece
	if (B->st->epsq == tsq && B->piece_on[fsq] == Pawn) {
		clear_bit(&occ, pawn_push(opp_color(stm), tsq));
		capture = Pawn;
	} else
		capture = B->piece_on[tsq];
	assert(capture != King);

	swap_list[0] = see_val[capture];
	clear_bit(&occ, fsq);
	attackers = calc_attackers(B, tsq, occ);

	// Handle promotion
	if (piece == Pawn && is_prank) {
		swap_list[0] += see_val[Queen] - see_val[Pawn];
		capture = Queen;
	} else
		capture = B->piece_on[fsq];

	// If the opponent has no attackers we are finished
	stm = opp_color(stm);
	stm_attackers = attackers & B->all[stm];
	if (!stm_attackers)
		return swap_list[0];

	/* The destination square is defended, which makes things rather more difficult to compute. We
	 * proceed by building up a "swap list" containing the material gain or loss at each stop in a
	 * sequence of captures to the destination square, where the sides alternately capture, and
	 * always capture with the least valuable piece. After each capture, we look for new X-ray
	 * attacks from behind the capturing piece. */
	do {
		/* Locate the least valuable attacker for the side to move. The loop below looks like it is
		 * potentially infinite, but it isn't. We know that the side to move still has at least one
		 * attacker left. */
		for (piece = Pawn; !(stm_attackers & B->b[stm][piece]); piece++)
			assert(piece < King);

		// remove the piece (from wherever it came)
		clear_bit(&occ, first_bit(stm_attackers & B->b[stm][piece]));
		// scan for new X-ray attacks through the vacated square
		attackers |= (get_all_RQ(B) & rook_attack(tsq, occ)) | (get_all_BQ(B) & bishop_attack(tsq, occ));
		// cut out pieces we've already done
		attackers &= occ;

		// add the new entry to the swap list (beware of promoting pawn captures)
		assert(sl_idx < 32);
		swap_list[sl_idx] = -swap_list[sl_idx - 1] + see_val[capture];
		if (piece == Pawn && is_prank) {
			swap_list[sl_idx] += see_val[Queen] - see_val[Pawn];
			capture = Queen;
		} else
			capture = piece;
		sl_idx++;

		stm = opp_color(stm);
		stm_attackers = attackers & B->all[stm];

		// Stop after a king capture
		if (piece == King && stm_attackers) {
			assert(sl_idx < 32);
			swap_list[sl_idx++] = see_val[King];
			break;
		}
	} while (stm_attackers);

	/* Having built the swap list, we negamax through it to find the best achievable score from the
	 * point of view of the side to move */
	while (--sl_idx)
		swap_list[sl_idx-1] = min(-swap_list[sl_idx], swap_list[sl_idx-1]);

	return swap_list[0];
}

int see_move(const Board *B, move_t m)
{
	assert(B->initialized);
	
	const unsigned * const see_val = Material[Opening];
	const unsigned fval = see_val[B->piece_on[m.fsq]];
	const unsigned tval = m.promotion == NoPiece
		? see_val[B->piece_on[m.tsq]] : see_val[m.promotion] - see_val[Pawn];

	if (tval > fval)
		return tval - fval;
	else
		return full_see(B, m.fsq, m.tsq);
}

bool board_is_draw(const Board *B)
{
	assert(B->initialized);
	
	// insufficient material
	if (!B->b[White][Pawn] && !B->b[Black][Pawn]
	        && B->st->npm[White] + B->st->npm[Black] <= max(MN, MB))
		return true;

	// 50 move rule
	if (B->st->rule50 >= 100)
		return true;

	// 3 move rule
	for (int i = 4; i <= B->st->rule50; i += 2)
		if (B->st[-i].key == B->st->key)
			return true;

	return false;
}

void set_square(Board *B, unsigned color, unsigned piece, unsigned sq, bool play)
{
	assert(B->initialized);
	assert(square_ok(sq) && color_ok(color) && piece_ok(piece) && B->piece_on[sq] == NoPiece);

	set_bit(&B->b[color][piece], sq);
	set_bit(&B->all[color], sq);
	B->piece_on[sq] = piece;

	if (play) {
		set_bit(&B->st->occ, sq);

		B->st->key ^= zobrist[color][piece][sq];
		if (piece == Pawn || piece == King)
			B->st->kpkey ^= zobrist[color][piece][sq];

		B->st->psq[color][Opening] += get_psqtab(color, Opening, piece, sq);
		B->st->psq[color][EndGame] += get_psqtab(color, EndGame, piece, sq);
		if (piece != Pawn)
			B->st->npm[color] += get_psqtab(color, Opening, piece, sq);
	}
}

void clear_square(Board *B, unsigned color, unsigned piece, unsigned sq, bool play)
{
	assert(B->initialized);
	assert(square_ok(sq) && color_ok(color) && piece_ok(piece) && B->piece_on[sq] == piece);

	clear_bit(&B->b[color][piece], sq);
	clear_bit(&B->all[color], sq);
	B->piece_on[sq] = NoPiece;

	if (play) {
		clear_bit(&B->st->occ, sq);

		B->st->key ^= zobrist[color][piece][sq];
		if (piece == Pawn || piece == King)
			B->st->kpkey ^= zobrist[color][piece][sq];

		B->st->psq[color][Opening] -= get_psqtab(color, Opening, piece, sq);
		B->st->psq[color][EndGame] -= get_psqtab(color, EndGame, piece, sq);
		if (piece != Pawn)
			B->st->npm[color] -= get_psqtab(color, Opening, piece, sq);
	}
}

uint64_t get_key(const Board *B)
{
	assert(B->initialized);
	return B->st->key
		^ (B->st->epsq == NoSquare ? 0ULL : zobrist_ep[B->st->epsq])
		^ zobrist_castle[B->st->crights];
}
