/*
 * DoubleCheck, a UCI chess playing engine. Copyright (C) 2011 Lucas Braesch.
 *
 * DoubleCheck is free software: you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * DoubleCheck is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not,
 * see <http://www.gnu.org/licenses/>.
*/
#pragma once
#include "bitboard.h"

typedef struct {
	unsigned piece, capture;	// undo info
	unsigned epsq;				// en passant square
	unsigned crights;			// castling rights, 4 bits in FEN order KQkq
	move_t last_move;			// last move played (for undo)
	uint64_t key, kpkey;		// zobrist keys: all and KP only
	uint64_t pinned, dcheckers;	// pinned and discovery check candidates from turn
	uint64_t attacked;			// squares attacked by opp_color(turn)
	uint64_t checkers;			// pieces checking turn's King
	uint64_t occ;				// occupancy
	int psq[NB_COLOR][NB_PHASE];// total material
	int npm[NB_COLOR];			// non pawn material
	unsigned rule50;			// counter for the 50 move rule
	uint64_t attacks[NB_COLOR][NB_PIECE+1];
} game_info;

typedef struct {
	uint64_t b[NB_COLOR][NB_PIECE];
	uint64_t all[NB_COLOR];
	unsigned turn;
	unsigned piece_on[NB_SQUARE];
	unsigned king_pos[NB_COLOR];
	game_info game_stack[MAX_GAME_PLY], *st;
	bool initialized;
} Board;

#define OO	1
#define OOO	2
static inline unsigned make_crights(unsigned color, unsigned cflag)
{
	assert(color_ok(color) && !(cflag & ~(OO | OOO)));
	return cflag << (2 * color);
}

void reset_board(Board *B);
void print_board(const Board *B);

#define MAX_FEN	80
void load_fen(Board *B, const char *fen);
void write_fen(const Board *B, char *fen);

void play(Board *B, move_t m);
void undo(Board *B);

bool board_is_draw(const Board *B);
int see_move(const Board *B, move_t m);

static inline unsigned color_on(const Board *B, unsigned sq)
{
	assert(B->initialized && square_ok(sq));
	return test_bit(B->all[White], sq) ? White : (test_bit(B->all[Black], sq) ? Black : NoColor);
}

static inline uint64_t get_RQ(const Board *B, unsigned color)
{
	assert(B->initialized && color_ok(color));
	return B->b[color][Rook] | B->b[color][Queen];
}

static inline uint64_t get_BQ(const Board *B, unsigned color)
{
	assert(B->initialized && color_ok(color));
	return B->b[color][Bishop] | B->b[color][Queen];
}

static inline uint64_t get_all_RQ(const Board *B)
{
	assert(B->initialized);
	return get_RQ(B, White) | get_RQ(B, Black);
}

static inline uint64_t get_all_BQ(const Board *B)
{
	assert(B->initialized);
	return get_BQ(B, White) | get_BQ(B, Black);
}

static inline uint64_t get_epsq_bb(const Board *B)
{
	assert(B->initialized);
	return B->st->epsq < NoSquare ? Bit[B->st->epsq] : 0ULL;
}

static inline bool board_is_check(const Board *B)
{
	assert(B->initialized);
	return B->st->checkers;
}

static inline bool board_is_double_check(const Board *B)
{
	assert(B->initialized);
	return !single_bit(B->st->checkers);
}

static inline bool has_pawn_on_7th(const Board *B, unsigned color)
{
	assert(B->initialized && color_ok(color));
	return B->b[color][Pawn] & PInitialRank[opp_color(color)];
}

void set_square(Board *B, unsigned color, unsigned piece, unsigned sq, bool play);
void clear_square(Board *B, unsigned color, unsigned piece, unsigned sq, bool play);

uint64_t get_key(const Board *B);