/*
 * DoubleCheck, a UCI chess playing engine. Copyright (C) 2011 Lucas Braesch.
 * 
 * DoubleCheck is free software: you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * DoubleCheck is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program. If not,
 * see <http://www.gnu.org/licenses/>.
*/
#include "eval.h"
#include "psq.h"

typedef struct {
	int16_t op[NB_COLOR], eg[NB_COLOR];
} eval_t;

static void add_eval(eval_t *e1, const eval_t *e2)
{
	assert(e1 && e2);
	e1->op[White] += e2->op[White];
	e1->op[Black] += e2->op[Black];
	e1->eg[White] += e2->eg[White];
	e1->eg[Black] += e2->eg[Black];
}

typedef struct {
	uint64_t kpkey;
	eval_t e;
	uint64_t passers;
} kp_hash_t;

#define PHT_SIZE	0x10000
static kp_hash_t PHT[PHT_SIZE];

typedef struct {
	unsigned sq, color;
	bool defended, exposed, isolated, backward, doubled, passed, unstoppable;
} PawnInfo;

static float calc_phase(const Board *B)
{
	unsigned rescale = 16*MP + 4*(MN+MB+MR) + 2*MQ;
	float material = (float)(B->st->npm[White] + B->st->npm[Black]) / rescale;
	return max(min(1-material, 1), 0);
}

static void eval_material(const Board *B, eval_t *e)
{
	if (!B->b[White][Pawn] && !B->b[Black][Pawn]) {
		const unsigned	// White piece count
			wn = count_bit(B->b[White][Knight]),
			wb = count_bit(B->b[White][Bishop]),
			wr = count_bit(B->b[White][Rook]),
			wq = count_bit(B->b[White][Queen]);
		const unsigned	// Black piece count
			bn = count_bit(B->b[Black][Knight]),
			bb = count_bit(B->b[Black][Bishop]),
			br = count_bit(B->b[Black][Rook]),
			bq = count_bit(B->b[Black][Queen]);
		const unsigned wm = wn + wb, bm = bn + bb;	// minor pieces
		bool drawish = false;
		
		if (wr == br && wq == bq	// equal majors
			&& wr + wq <= 1) {		// but no more than 1
			if (abs(wm - bm) <= 1)
				drawish = true;		// equal minors +/- 1
			else if ((2 == wn && bm <= 2) || (2 == bn && wm <= 2))
				drawish = true;		// 2 Knights vs 2 minors or less
		}
		else if (wq == bq && wr == br+1 && (bm == wm+1 || bm == wm+2))
			drawish = true;	// Rook advantage for white, compensated by 1 or 2 minors
		else if (wq == bq && br == wr+1 && (wm == bm+1 || wm == bm+2))
			drawish = true;	// Rook advantage for black, compensated by 1 or 2 minors
		else if (wq == bq+1 && br == wr+2 && wm == bm)
			drawish = true;	// Queen advantage for white, compensated by 2 Rooks
		else if (bq == wq+1 && wr == br+2 && wm == bm)
			drawish = true;	// Queen advantage for black, compensated by 2 Rooks
			
		if (drawish) {
			for (unsigned color = White; color <= Black; color++) {
				e->op[color] = get_psqtab(color, Opening, King, B->king_pos[color]);
				e->eg[color] = get_psqtab(color, EndGame, King, B->king_pos[color]);
			}
			return;
		}		
	}

	for (unsigned color = White; color <= Black; color++) {
		e->op[color] += B->st->psq[color][Opening];
		e->eg[color] += B->st->psq[color][EndGame];

		bool bpair = count_bit(B->b[color][Bishop]) >= 2;
		e->op[color] += BP * bpair;
		e->eg[color] += BP * bpair;
	}
}

static void eval_hanging(const Board *B, eval_t *e)
/* Penalise hanging pieces */
{
	for (unsigned color = White; color <= Black; color++) {
		const unsigned us = color, them = opp_color(us);
		
		// loose pawns = undefended pawns
		uint64_t loose_pawns = B->b[us][Pawn] & ~B->st->attacks[us][NoPiece];
		
		// loose pieces = pieces either attacked by a pawn or not defended by a pawn
		uint64_t loose_pieces = (B->all[us] & ~B->b[us][Pawn])
			& (B->st->attacks[them][Pawn] | ~B->st->attacks[us][Pawn]);
		
		// hanging = loose and attacked
		uint64_t hanging = (loose_pawns | loose_pieces) & B->st->attacks[them][NoPiece];
		
		// scoring (a bit simplistic at the moment, at least it's fast)
		while (hanging) {
			unsigned sq = next_bit(&hanging), victim = B->piece_on[sq];
			e->op[us] -= 5 + Material[Opening][victim]/32;
			e->eg[us] -= 10 + Material[EndGame][victim]/32;
		}
	}
}

static void eval_mobility(const Board *B, eval_t *e)
{
	static const int mob_zero[NB_PIECE] = {0, 3, 4, 5, 0, 0};
	static const unsigned mob_unit[NB_PHASE][NB_PIECE] = {
		{0, 4, 5, 2, 1, 0},		// Opening
		{0, 4, 5, 4, 2, 0}		// EndGame
	};

	for (unsigned color = White; color <= Black; color++) {
		const unsigned us = color, them = opp_color(us);
		const uint64_t mob_targets = ~(B->st->attacks[them][Pawn] | B->b[us][Pawn]);
		
		uint64_t fss, tss, occ;
		unsigned fsq, piece;
		int count;
		
		/* Generic linear mobility */
		#define MOBILITY(p0, p)								\
			count = count_bit_max15(tss) - mob_zero[p0];	\
			e->op[us] += count * mob_unit[Opening][p];		\
			e->eg[us] += count * mob_unit[EndGame][p]
		
		/* Knight mobility */
		fss = B->b[us][Knight];
		while (fss) {
			fsq = next_bit(&fss);
			tss = NAttacks[fsq] & mob_targets;
			MOBILITY(Knight, Knight);
		}
		
		/* Lateral mobility */
		fss = get_RQ(B, us);
		occ = B->st->occ & ~B->b[us][Rook];		// see through rooks
		while (fss) {
			fsq = next_bit(&fss); piece = B->piece_on[fsq];
			tss = rook_attack(fsq, occ) & mob_targets;
			MOBILITY(Rook, piece);
		}

		/* Diagonal mobility */
		fss = get_BQ(B, us);
		occ = B->st->occ & ~B->b[us][Bishop];	// see through bishops
		while (fss) {
			fsq = next_bit(&fss); piece = B->piece_on[fsq];
			tss = bishop_attack(fsq, occ) & mob_targets;
			MOBILITY(Bishop, piece);
		}
	}
}

static void eval_pieces(const Board *B, eval_t *e)
{
	static const unsigned Rook7th = 40, RookOpenFile = 16;
	static const uint64_t OutPostZone[NB_COLOR] = {
		0x00003C3C3C000000ULL,	// Rank 4->6, File C->F
		0x0000003C3C3C0000ULL	// Rank 3->5, File C->F
	};
	
	for (unsigned color = White; color <= Black; color++) {
		const unsigned us = color, them = opp_color(us);
		const uint64_t our_pawns = B->b[us][Pawn], their_pawns = B->b[them][Pawn];
		uint64_t fss, tss;
		unsigned count;
		
		/* Rook on (semi)open file */
		fss = B->b[us][Rook];
		while (fss) {
			const unsigned fsq = next_bit(&fss);
			const uint64_t tss = file_bb(fsq);
			unsigned bonus = RookOpenFile;
			
			if (!(tss & our_pawns)) {			// semi open file ?
				if (tss & B->b[them][King])
					bonus += bonus/2;			// increase bonus on king file
				
				if (!(tss & their_pawns)) {		// open file ?
					e->op[us] += bonus;
					e->eg[us] += bonus/2;
				} else {						// no, just semi open
					e->op[us] += bonus/2;
					e->eg[us] += bonus/4;
				}
			}
		}
		
		/* Rook or Queen on 7th rank */
		fss = get_RQ(B, us);
		tss = PInitialRank[them];
		if (	(fss & tss)									// rook or queen on 7th ?
			&&	((PPromotionRank[us] & B->b[them][King])	// enemy king on 8th ?
				||	(tss & B->b[them][Pawn]))				// or enemy pawns on 7th ?
			)
		{
			if ((count = count_bit(B->b[us][Rook] & tss))) {	// rook on 7th ?
				e->op[us] += count * Rook7th/2;
				e->eg[us] += count * Rook7th;
			}
			if ((count = count_bit(B->b[us][Queen] & tss))) {	// queen on 7th ?
				e->op[us] += count * Rook7th/4;
				e->eg[us] += count * Rook7th/2;
			}			
		}
		
		/* Knight outposts */
		fss = B->b[us][Knight] & OutPostZone[us] & ~B->st->attacks[them][Pawn];
		if ((count = count_bit(fss))) {
			e->op[us] += count * 2;
			e->eg[us] += count * 2;
			
			if ((count = count_bit(fss & B->st->attacks[us][Pawn]))) {
				e->op[us] += count * 3;
				e->eg[us] += count * 3;
			}
		}

		/* Bishop outposts */
		fss = B->b[us][Bishop] & OutPostZone[us] & ~B->st->attacks[them][Pawn];
		if ((count = count_bit(fss))) {
			e->op[us] += count;
			e->eg[us] += count;
			
			if ((count = count_bit(fss & B->st->attacks[us][Pawn]))) {
				e->op[us] += count * 2;
				e->eg[us] += count * 2;
			}
		}
	}
}

/* King safety test positions:
 * 4r1k1/2q1pp2/3n2pQ/2rPR3/pp4P1/1P5R/1P2NP2/6K1 b - - 0 1
 * 2R3Q1/5p1p/4p1p1/1P1b3k/3Pp2P/6P1/r4qBK/8 w - - 0 46
 * 1R2Rbk1/8/3q2p1/5pNp/p6P/5P1K/6P1/8 w - - 0 71
 * 1r1r1bk1/5p2/2p1pQBp/p2qP3/Pp1P1P2/4P2R/1P4RP/7K b - - 0 29
 * 3rr3/1p3k2/pNbpp3/P1p5/2P3Q1/1qPK1P1P/4R1P1/4R3 w - - 0 47
 * 4r1k1/2q1pp2/3n2pQ/2rPR1N1/pp4P1/1P5R/1P3P2/6K1 b - -
 * 2k4r/2b1Rpp1/p1Qp3p/8/8/8/qPP2PPP/2B3K1 b - - 0 25
 * 2r4k/4b1p1/1r1p1pPp/p3pP1P/4P3/6R1/NPP5/1K3B2 w - - 0 38
 * 6R1/5R2/2b4k/P3q2p/6pP/1B6/P1K5/8 w - - 0 1
 * 3RRbk1/5r1p/5rpB/pp2p3/2n1P3/2P2P1P/1q2N1QK/8 b - - 0 1
 * 2r2rk1/1bq2pp1/p2p1P2/1p2p1n1/4Pb2/PNNQ4/1PP5/1K1B2RR b - - 0 1
 * 5R2/r5k1/3p2b1/p1pB4/P1P2R1p/1PP3p1/4r2p/5N1K b - - 0 1
 * r2r2k1/6pp/b2pp1p1/q7/1p1BPP1P/pP1P1Q2/P7/1KR4R w - - 0 1
 * 2k1r3/pppbrppp/3b4/8/3Pq3/2P4P/PP2B3/R1BQKRN1 b - - 0 1
 * 6kr/4pp2/5q2/2rPR1p1/pp4Pn/1P6/1P1N1P1Q/6KR b - - 0 1
 * 1r1n2k1/3bqpnp/1r2p3/p2p4/3P4/3B2Q1/PP1R1P1P/K5R1 w - - 0 33
 * 3r4/p1pb1rpk/7p/3P1p1P/P1P2Qnq/2B1P1N1/2P1R1P1/4R1K1 w - - 0 29
 * 2r5/1q3p2/4p1k1/np1pPpPp/p2BrP1P/P1P2Q2/1P4R1/1K4R1 b - - 0 67
 * 8/5pk1/3R2p1/4n2p/4P2P/qrr2PP1/5RK1/3Q1N2 b - - 0 51
 * 2r5/1q3p2/4p1k1/np1pPpPp/p2BrP1P/P1P3R1/1P4Q1/1K4R1 b - - 0 18
 * 1prr3k/p6b/1Q3p2/P3q2P/2P1P1R1/5PP1/1R4K1 w - - 0 40
 * 1R2bk2/q1r2pp1/r1p4p/p1R1PP2/P4QPP/1P2B3/1b4B1/5K2 b - - 0 14
 * 1k1r3r/pp3p2/1np1pp2/8/PPPPq1pP/2Q3P1/5P2/R2R1BK1 b - - 0 24
 * 4rn2/4p3/5qk1/2rPR3/pp4P1/1P6/1P1N1P1Q/6KR b - - 0 1
 * r3br2/2pq1N1k/1p1p4/1P1Pp1Qn/p1P1P3/4BP2/1P1N4/1K4R1 b - - 0 41
 * 2rq1k2/Q2R1p1p/p2p1rp1/1p1P4/5P2/2P5/PP5P/1K2R3 b - - 0 33
*/
static void eval_safety(const Board *B, eval_t *e)
{
	//static const unsigned AttackWeight[NB_PIECE] = {1, 2, 2, 3, 0, 0};
	static const unsigned AttackWeight[NB_PIECE] = {2, 3, 3, 4, 0, 0};
	static const unsigned ShieldWeight = 3;
	
	for (unsigned color = White; color <= Black; color++) {
		const unsigned us = color, them = opp_color(us), ksq = B->king_pos[us];
		
		// General attack zone
		const uint64_t attack_zone = KAttacks[ksq]
			& ~(B->st->attacks[us][Pawn] | B->b[them][Pawn]);
		// Knight attack zone
		const uint64_t knight_attack_zone = (KAttacks[ksq] | NAttacks[ksq])
			& ~(B->st->attacks[us][Pawn] | B->b[them][Pawn]);
		// Defended by our pieces
		const uint64_t defended = B->st->attacks[us][Knight]
			| B->st->attacks[us][Bishop]
			| B->st->attacks[us][Rook]
			| B->st->attacks[us][Queen];

		unsigned total_weight = 0, total_count = 0, sq, count;
		uint64_t sq_attackers, attacked, occ, fss;

		// Generic attack scoring
		#define ADD_ATTACK(p0)								\
			if (sq_attackers) {								\
				count = count_bit(sq_attackers);			\
				total_weight += AttackWeight[p0] * count;	\
				if (test_bit(defended, sq)) count--;		\
				total_count += count;						\
			}

		// Knight attacks
		attacked = B->st->attacks[them][Knight] & knight_attack_zone;
		fss = B->b[them][Knight];
		while (attacked) {
			sq = next_bit(&attacked);
			sq_attackers = NAttacks[sq] & fss;
			ADD_ATTACK(Knight);
		}

		// Lateral attacks
		attacked = (B->st->attacks[them][Rook] | B->st->attacks[them][Queen]) & attack_zone;
		fss = get_RQ(B, them);
		occ = B->st->occ & ~fss;	// rooks and queens see through each other
		while (attacked) {
			sq = next_bit(&attacked);
			sq_attackers = fss & RPseudoAttacks[sq] & rook_attack(sq, occ);
			ADD_ATTACK(Rook);
		}

		// Diagonal attacks
		attacked = (B->st->attacks[them][Bishop] | B->st->attacks[them][Queen]) & attack_zone;
		fss = get_BQ(B, them);
		occ = B->st->occ & ~fss;	// bishops and queens see through each other
		while (attacked) {
			sq = next_bit(&attacked);
			sq_attackers = fss & BPseudoAttacks[sq] & bishop_attack(sq, occ);
			ADD_ATTACK(Bishop);
		}
		
		// Pawn attacks
		fss = PawnsAttacking[us][ksq] & B->b[them][Pawn];
		while (fss) {
			sq = next_bit(&fss);
			total_count += count = count_bit(PAttacks[them][sq] & attack_zone);
			total_weight += AttackWeight[Pawn] * count;
		}
		
		// Adjust for king's "distance to safety"
		total_count += min(king_dist(ksq, us ? G8 : G1), king_dist(ksq, us ? B8 : B1));
		
		// Adjust for lack of pawn shield
		total_count += count = 3 - count_bit(Shield[us][ksq] & B->b[us][Pawn]);
		total_weight += count * ShieldWeight;
		
		e->op[us] -= total_weight * total_count;
	}
}

static void do_eval_pawns(const Board *B, eval_t *ep, uint64_t *passers)
{
	// penalties
	static const unsigned DoubledOpening = 10, DoubledEndgame = 20;
	static const unsigned IsolatedOpening = 10, IsolatedOpeningOpen = 20, IsolatedEndgame = 20;
	static const unsigned BackwardOpening = 8, BackwardOpeningOpen = 16, BackwardEndgame = 10;
	// bonuses
	static const unsigned Chain = 5, Storm = 3;
	
	for (unsigned color = White; color <= Black; color++) {
		const unsigned us = color, them = opp_color(us);
		const uint64_t our_pawns = B->b[us][Pawn], their_pawns = B->b[them][Pawn];
		
		uint64_t pawns = our_pawns;
		while (pawns) {
			const unsigned sq = next_bit(&pawns), r = rank(sq);
			const uint64_t besides = our_pawns & adjacent_files_bb(sq);
			bool doubled, isolated, backward, chain, passed, open, storm, candidate;
			
			isolated = !besides;
			doubled = SquaresInFront[us][sq] & our_pawns;	// doubled = friendly pawn in front
			open = !doubled && !(SquaresInFront[us][sq] & their_pawns);	// open = no pawn in front
			chain = besides & (Rank[r] | Rank[us ? r+1 : r-1]);
			passed = open && !(PassedPawnMask[us][sq] & their_pawns);
			backward = !(passed | isolated | chain)
				&& !(InFront[them][r] & besides)		// no friendly pawn behind on adjacent files
				&& !(PAttacks[us][sq] & their_pawns)	// can't capture an enemy pawn
				&& (PAttacks[us][pawn_push(us, sq)] & their_pawns);	// can't push
			storm = !doubled && test_bit(PassedPawnMask[them][B->king_pos[them]] & HalfBoard[them], sq);
			candidate = !passed && chain && !isolated && open
				&& single_bit(PawnSpanMask[us][sq] & their_pawns);
			
			if (doubled) {
				ep->op[us] -= DoubledOpening;
				ep->eg[us] -= DoubledEndgame;
			}
			
			if (isolated) {
				ep->op[us] -= open ? IsolatedOpeningOpen : IsolatedOpening;
				ep->eg[us] -= IsolatedEndgame;
			}
			
			if (backward) {
				ep->op[us] -= open ? BackwardOpeningOpen : BackwardOpening;
				ep->eg[us] -= BackwardEndgame;
			}
			
			if (chain)
				ep->op[us] += Chain;
			
			if (storm)
				ep->op[us] += Storm;
			
			if (candidate) {
				int n = us ? 7-r : r;
				const int d1 = king_dist(sq, B->king_pos[us]), d2 = king_dist(sq, B->king_pos[them]);
				if (d1 > d2) n -= d1 - d2;	// penalise if enemy king is closer
				n = max(Rank1, n);
				ep->eg[us] += n*n;
			} else if (passed) {
				set_bit(passers, sq);
				const int L = (us ? 7-r : r)-Rank2;		// Linear part		0..5
				const int Q = L*(L-1);					// Quadratic part	0..20

				//  score based on rank
				ep->op[us] += 8 * Q;
				ep->eg[us] += 4 * (Q + L + 1);
				
				if (Q) {
					//  adjustment for king distance
					const int next_sq = pawn_push(us, sq);
					ep->eg[us] += king_dist(next_sq, B->king_pos[them]) * 2 * Q;
					ep->eg[us] -= king_dist(next_sq, B->king_pos[us]) * Q;
					if (rank(next_sq) != (us ? Rank1 : Rank8))
						ep->eg[us] -= king_dist(pawn_push(us, next_sq), B->king_pos[us]) * Q / 2;
				}
				
				// support by friendly pawn
				if (chain)
					ep->eg[us] += 6 * L;
			}
		}
	}
}

static void eval_pawns(const Board *B, eval_t *e)
{
	static bool first_call = true;
	if (first_call) {
		memset(PHT, 0, sizeof(PHT));
		first_call = false;
	}
	
	const uint64_t kpkey = B->st->kpkey;
	const unsigned idx = kpkey & (PHT_SIZE-1);
	kp_hash_t * const kph = &PHT[idx];
	
	if (kph->kpkey != kpkey) {
		memset(kph, 0, sizeof(kp_hash_t));
		do_eval_pawns(B, &kph->e, &kph->passers);
		kph->kpkey = kpkey;
	}
	add_eval(e, &kph->e);
	
	// piece-dependant passed pawn scoring
	uint64_t b = kph->passers;
	while (b) {
		const unsigned sq = next_bit(&b);
		const unsigned us = color_on(B, sq), them = opp_color(us);
		
		if (B->all[them] == (B->b[them][Pawn] | B->b[them][King])) {
			// opponent has no pieces
			const unsigned psq = square(us ? Rank1 : Rank8, file(sq));
			const unsigned pdist = king_dist(sq, psq);
			int kdist = king_dist(B->king_pos[them], psq) - (them == B->turn);
			
			if (kdist > pdist) {	// unstoppable passer
				e->eg[us] += MQ - (pdist+1)*EP;
				continue;			// do not cummulate with normal passed pawns terms below
			}
		}
		
		const int r = rank(sq);
		const int L = (us ? 7-r : r)-Rank2;		// Linear part		0..5
		const int Q = L*(L-1);					// Quadratic part	0..20
		const int next_sq = pawn_push(us, sq);
			
		if (Q && !test_bit(B->st->occ, next_sq)) {
			const uint64_t path = in_front_bb(us, sq);
			const uint64_t b = file_bb(sq) & rook_attack(sq, B->st->occ);

			uint64_t defended, attacked;
			if (get_RQ(B, them) & b) {
				defended = path & B->st->attacks[us][NoPiece];
				attacked = path;
			} else {
				defended = (get_RQ(B, us) & b) ? path : path & B->st->attacks[us][NoPiece];
				attacked = path & (B->st->attacks[them][NoPiece] | B->all[them]);
			}
			
			if (!attacked)
				e->eg[us] += Q * (path == defended ? 7 : 6);
			else
				e->eg[us] += Q * ((attacked & defended) == attacked ? 5 : 3);
		}
	}
}

static int interpolate(const Board *B, eval_t *e)
{
	const unsigned us = B->turn, them = opp_color(us);
	const float phase = calc_phase(B);
	
	// tempo bonus
	e->op[us] += 9;
	e->eg[us] += 4;
	
	return (1-phase)*(e->op[us]-e->op[them]) + phase*(e->eg[us]-e->eg[them]);
}

int calc_eval(const Board *B)
{
	assert(B->initialized && !board_is_check(B));
	
	eval_t e;
	memset(&e, 0, sizeof(eval_t));
	
	eval_material(B, &e);
	eval_mobility(B, &e);
	eval_pieces(B, &e);
	eval_pawns(B, &e);
	eval_hanging(B, &e);
	eval_safety(B, &e);

	return interpolate(B, &e);
}

static void print_component(const char *name, eval_t *e)
{
	printf("%s = W(%d, %d) B(%d, %d)\n", name,
		e->op[White], e->eg[White],
		e->op[Black], e->eg[Black]);
}

void print_eval(const Board *B)
/* Eval breakdown, by component
 * TODO: figure out how to pretty print this in a table, a la Stockfish */
{
	assert(!board_is_check(B));
	
	eval_t component[6];
	memset(&component, 0, sizeof(component));
	
	eval_material(B, &component[0]);
	eval_mobility(B, &component[1]);
	eval_pieces(B, &component[2]);
	eval_pawns(B, &component[3]);
	eval_hanging(B, &component[4]);
	eval_safety(B, &component[5]);
	
	print_component("material", &component[0]);
	print_component("mobility", &component[1]);
	print_component("pieces", &component[2]);
	print_component("pawns", &component[3]);
	print_component("loose", &component[4]);
	print_component("safety", &component[5]);

	eval_t total;
	memset(&total, 0, sizeof(total));
	for (unsigned i = 0; i < sizeof(component)/sizeof(eval_t); add_eval(&total, &component[i++]));
	
	const int result = interpolate(B, &total);
	assert(result == calc_eval(B));

	print_component("total", &total);
	printf("=> eval: %d\n", result);
}
