/*
 * DoubleCheck, a UCI chess playing engine. Copyright (C) 2011 Lucas Braesch.
 * 
 * DoubleCheck is free software: you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * DoubleCheck is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program. If not,
 * see <http://www.gnu.org/licenses/>.
*/
#include "move.h"
#include "board.h"
#include "movegen.h"

bool move_ok(const Board *B, move_t m)
/* Determines whether a move is legal in the current board position. This function makes no
 * assumption or shortcut (like assuming the move is at least pseudo-legal for example), and tests
 * every single component of the move_t structure. So it is quite long and complex, but its execution
 * time should be rather fast (with the notable exception of check evasions)
 * */
{
	const unsigned us = B->turn, them = opp_color(us), kpos = B->king_pos[us];
	const unsigned piece = B->piece_on[m.fsq], capture = B->piece_on[m.tsq];
	const unsigned fsq = m.fsq, tsq = m.tsq;
	
	// verify fsq and piece
	if (color_on(B, fsq) != us)
		return false;
	
	// verify tsq and capture
	if (color_on(B, tsq) != (capture == NoPiece ? NoColor : them))
		return false;

	// check evasion: use the slow method (optimize later if needs be)
	if (board_is_check(B)) {
		move_t mlist[MAX_MOVES];
		move_t * const end = gen_evasion(B, mlist);

		for (move_t *m_ptr = mlist; m_ptr < end; m_ptr++)
			if (move_equal(*m_ptr, m))
				return true;

		return false;
	}

	bool pinned = test_bit(B->st->pinned, fsq);
	uint64_t pin_ray = Direction[kpos][fsq];
	
	// pawn moves
	if (piece == Pawn) {
		// generic self check
		if (pinned && !test_bit(pin_ray, tsq))
			return false;
			
		// m.promotion is a valid promotion piece <==> tsq is on the 8th rank
		if ((bool)test_bit(PPromotionRank[us], tsq) != (Knight <= m.promotion && m.promotion <= Queen))
			return false;

		const int delta = tsq - fsq;
		const bool notA = file(fsq) != FileA, notH = file(fsq) != FileH;
		const bool delta_capture = us
			? ((notA && -9 == delta) || (notH && -7 == delta))
			: ((notA && +7 == delta) || (notH && +9 == delta));
		const bool delta_push = (us ? -8 : +8) == delta;
		const bool delta_dpush = (us ? -16 : +16) == delta;

		if (m.ep) {
			// must be a capture delta and land on the epsq
			if (B->st->epsq != tsq || !delta_capture)
				return false;
			
			// self check through en passant captured pawn
			const unsigned ep_cap = pawn_push(them, tsq);
			const uint64_t ray = Direction[kpos][ep_cap];
			if (ray) {
				uint64_t occ = B->st->occ;
				clear_bit(&occ, ep_cap);
				set_bit(&occ, tsq);
				
				const uint64_t RQ = get_RQ(B, them);
				if (ray & RQ & RPseudoAttacks[kpos] & rook_attack(kpos, occ))
					return false;
				
				const uint64_t BQ = get_BQ(B, them);
				if (ray & BQ & BPseudoAttacks[kpos] & bishop_attack(kpos, occ))
					return false;
			}
			
			return true;
		}
		
		if (delta_capture)
			return capture != NoPiece;
		else if (delta_push)
			return capture == NoPiece;
		else if (delta_dpush)
			return capture == NoPiece
				&& B->piece_on[fsq+delta/2] == NoPiece
				&& test_bit(PInitialRank[us], fsq);
		else
			return false;
	}

	if (m.ep || m.promotion != NoPiece)
		return false;
		
	uint64_t tss = piece_attack(piece, fsq, B->st->occ);
	
	if (piece == King) {
		// castling moves
		if (is_castling(B, m)) {
			uint64_t safe /* must not be attacked */, empty /* must be empty */;
			
			if ((fsq+2 == tsq) && (B->st->crights & make_crights(us, OO))) {
				empty = safe = Bit[m.fsq+1] | Bit[m.fsq+2];
				return !(B->st->attacked & safe) && !(B->st->occ & empty);
			} else if ((fsq-2 == tsq) && (B->st->crights & make_crights(us, OOO))) {
				safe = Bit[m.fsq-1] | Bit[m.fsq-2];
				empty = safe | Bit[m.fsq-3];
				return !(B->st->attacked & safe) && !(B->st->occ & empty);
			} else
				return false;
		}
		
		// normal king moves
		return test_bit(tss & ~B->st->attacked, tsq);
	}

	// piece moves
	if (pinned) tss &= pin_ray;
	return test_bit(tss, tsq);
}

move_t string_to_move(const Board *B, const char *s)
{
	assert(B->initialized && s);
	move_t m;

	m.fsq = square(s[1]-'1', s[0]-'a');
	m.tsq = square(s[3]-'1', s[2]-'a');
	m.ep = B->piece_on[m.fsq] == Pawn && m.tsq == B->st->epsq;

	if (s[4]) {
		unsigned piece;
		for (piece = Knight; piece <= Queen; piece++)
			if (PieceLabel[Black][piece] == s[4])
				break;
		assert(piece < King);
		m.promotion = piece;
	} else
		m.promotion = NoPiece;

	return m;
}

void move_to_string(const Board *B, move_t m, char *s, bool pretty)
{
	assert(B->initialized && s);
	
	if (pretty) {
		const unsigned check = is_check(B, m);
		
		if (is_OO(B, m)) {
			strcpy(s, "OO");
			s += 2;
			if (check)
				*s++ = '+';
		} else if (is_OOO(B, m)) {
			strcpy(s, "OOO");
			s += 3;
			if (check)
				*s++ = '+';
		} else {
			const unsigned piece = B->piece_on[m.fsq], capture = B->piece_on[m.tsq];
			if (piece != Pawn)
				*s++ = PieceLabel[White][piece];

			assert(square_ok(m.fsq));
			*s++ = file(m.fsq) + 'a';
			*s++ = rank(m.fsq) + '1';

			if (capture < NoPiece || m.ep)
				*s++ = 'x';

			*s++ = file(m.tsq) + 'a';
			*s++ = rank(m.tsq) + '1';
			assert(square_ok(m.tsq));

			if (m.ep) {
				strcpy(s, "/ep");
				s += 3;
			} else if (piece_ok(m.promotion)) {
				*s++ = '=';
				*s++ = PieceLabel[White][m.promotion];
			}

			if (check)
				*s++ = '+';
		}
	} else {
		*s++ = file(m.fsq) + 'a';
		*s++ = rank(m.fsq) + '1';
		assert(square_ok(m.fsq));

		*s++ = file(m.tsq) + 'a';
		*s++ = rank(m.tsq) + '1';
		assert(square_ok(m.tsq));

		if (piece_ok(m.promotion))
			*s++ = PieceLabel[Black][m.promotion];
	}

	*s++ = '\0';
}

unsigned is_check(const Board *B, move_t m)
/* Tests if a move is checking the enemy king. General case: direct checks, revealed checks (when
 * piece is a dchecker moving out of ray). Special cases: castling (check by the rook), en passant
 * (check through a newly revealed sliding attacker, once the ep capture square has been vacated)
 * returns 2 for a discovered check, 1 for any other check, 0 otherwise */
{
	assert(B->initialized);
	const unsigned us = B->turn, them = opp_color(us), kpos = B->king_pos[them];

	// test discovered check
	if ((test_bit(B->st->dcheckers, m.fsq))		// discovery checker
		&& (!test_bit(Direction[kpos][m.fsq], m.tsq)))	// move out of its dc-ray
		return 2;
	// test direct check
	else if (m.promotion == NoPiece) {
		const unsigned piece = B->piece_on[m.fsq];
		uint64_t tss = piece == Pawn ? PAttacks[us][m.tsq]
			: piece_attack(piece, m.tsq, B->st->occ);
		if (test_bit(tss, kpos))
			return 1;
	}

	if (m.ep) {
		uint64_t occ = B->st->occ;
		// play the ep capture on occ
		clear_bit(&occ, m.fsq);
		clear_bit(&occ, m.tsq + (us ? 8 : -8));
		set_bit(&occ, m.tsq);
		// test for new sliding attackers to the enemy king
		if ((get_RQ(B, us) & RPseudoAttacks[kpos] & rook_attack(kpos, occ))
		|| (get_BQ(B, us) & BPseudoAttacks[kpos] & bishop_attack(kpos, occ)))
			return 2;	// discovered check through the fsq or the ep captured square
	}
	else if (is_castling(B, m)) {
		// position of the Rook after playing the castling move
		unsigned rook_sq = is_OO(B, m) ? m.tsq - 1 : m.tsq + 1;
		uint64_t occ = B->st->occ;
		clear_bit(&occ, m.fsq);
		uint64_t RQ = get_RQ(B, us);
		set_bit(&RQ, rook_sq);
		if (RQ & RPseudoAttacks[kpos] & rook_attack(kpos, occ))
			return 1;	// direct check by the castled rook
	}
	else if (m.promotion < NoPiece) {
		// test check by the promoted piece
		uint64_t occ = B->st->occ;
		clear_bit(&occ, m.fsq);
		if (test_bit(piece_attack(m.promotion, m.tsq, occ), kpos))
			return 1;	// direct check by the promoted piece
	}

	return 0;
}