/*
 * DoubleCheck, a UCI chess playing engine. Copyright (C) 2011 Lucas Braesch.
 * 
 * DoubleCheck is free software: you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * DoubleCheck is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program. If not,
 * see <http://www.gnu.org/licenses/>.
*/
#include "movesort.h"

static int History[NB_COLOR][NB_PIECE][NB_SQUARE];
static const int HistoryMax = 2000;

void clear_history()
{
	memset(History, 0, sizeof(History));
}

int get_history(const Board *B, move_t m)
{
	assert(!is_cop(B, m));
	return History[B->turn][B->piece_on[m.fsq]][m.tsq];
}

void history_update(const Board *B, move_t m, int bonus)
{
	assert(!is_cop(B, m));
	int *h = &History[B->turn][B->piece_on[m.fsq]][m.tsq];
	*h += bonus;

	if (*h > HistoryMax)
		*h = HistoryMax;
	else if (*h < -HistoryMax)
		*h = -HistoryMax;
}

static int mvv_lva(const Board *B, move_t m)
{
	// Queen is the best capture available (King can't be captured since move is legal)
	static const unsigned victim_value[NB_PIECE+1] = {1, 2, 2, 3, 4, 0, 0};
	// King is the best attacker (since move is legal) followed by Pawn etc.
	static const unsigned attacker_value[NB_PIECE] = {4, 3, 3, 2, 1, 5};

	const unsigned piece = B->piece_on[m.fsq], capture = B->piece_on[m.tsq];
	return (victim_value[capture] + victim_value[m.promotion]) * 8 + attacker_value[piece];
}

void gen_msl(const Board *B, MoveSortList *msl, int gen_what, const SearchInfo *si)
/* values of gen_what:
 * 0 all legal moves (search depth > 0)
 * 1 captures, quiet promotions, quiet piece checks (qsearch at depth 0)
 * 2 captures, quiet promotions (qsearch at depth < 0)
 * */
{
	/* generate moves */

	move_t ml[MAX_MOVES];
	move_t *m = ml, *end;

	if (board_is_check(B))
		end = gen_evasion(B, m);
	else {
		const uint64_t
		friends = B->all[B->turn],
		enemies = B->all[opp_color(B->turn)];
		uint64_t targets;

		targets = gen_what ? enemies : ~friends;
		m = gen_piece_moves(B, targets, m, true);

		targets = gen_what ? (enemies | PPromotionRank[B->turn]) & ~friends : ~friends;
		m = gen_pawn_moves(B, targets, m, gen_what == 0);

		if (gen_what == 0)
			m = gen_castling(B, m);
		else if (gen_what == 1) {
			m = gen_castling(B, m);
			m = gen_quiet_checks(B, m);
		}
		
		end = m;
	}

	/* fill the MoveSortList msl */

	msl->count = end - ml;
	msl->idx = 0;

	for (int i = 0; i < msl->count; i++) {
		MoveSort *ms = &msl->ms[i];
		ms->m = ml[i];
		ms->see = -INF;		// uninitialized value

		// 1. TT move
		if (move_equal(ms->m, si->tt))
			ms->score = INF * HistoryMax;
		// quiet moves
		else if (!is_cop(B, ms->m)) {
			// killer moves
			if (move_equal(ms->m, si->killer[0]))
				ms->score = HistoryMax - 1;
			else if (move_equal(ms->m, si->killer[1]))
				ms->score = HistoryMax - 2;
			// other quiet moves by history score
			else
				ms->score = get_history(B, ms->m);
		}
		// captures by see or mvv/lva
		else {
			if (gen_what == 2 && !board_is_check(B)) {
				// only captures and promotions at this node, use mvv/lva
				ms->score = mvv_lva(B, ml[i]);
			}
			else {
				// mix SEE and History scores
				ms->see = see_move(B, ml[i]);
				ms->score = HistoryMax * ms->see / MP;
			}
		}
	}
}

MoveSort *next_move(const Board *B, MoveSortList *msl)
{
	const int i0 = msl->idx;
	if (i0 >= msl->count)
		return NULL;

	for (int i = i0; i < msl->count; i++)
		if (msl->ms[i].score > msl->ms[i0].score) {
			MoveSort temp = msl->ms[i0];
			msl->ms[i0] = msl->ms[i];
			msl->ms[i] = temp;
		}

	msl->idx++;
	return &msl->ms[i0];
}