/*
 * DoubleCheck, a UCI chess playing engine. Copyright (C) 2011 Lucas Braesch.
 * 
 * DoubleCheck is free software: you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * DoubleCheck is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program. If not,
 * see <http://www.gnu.org/licenses/>.
 * 
 * Credits: Parametric PSQ generation based on Fruit 2.1. Note however, that the code was completely
 * rewritten, and a few things have been siplified. I'm basically trying to minimize the number of
 * arbitrary parameters, with a view to doing a brute force CLOP optimisation at a later stage.
*/
#include "psq.h"

int PSQTable[NB_PHASE][NB_PIECE][NB_SQUARE];
bool PSQInitialized = false;

/* Shape */
static const int Center[8]		= {-3, -1, +0, +1, +1, +0, -1, -3};
static const int KnightRank[8]	= {-2, -1, +0, +1, +2, +3, +2, +1};
static const int KingFile[8]	= {+3, +4, +2, +0, +0, +2, +4, +3};
static const int KingRank[8]	= {+1, +0, -2, -3, -4, -5, -6, -7};

/* Weight */
static const int PawnFileOpening = 5;
static const int KnightCentreOpening = 5;
static const int KnightCentreEndgame = 5;
static const int KnightRankOpening = 5;
static const int BishopCentreOpening = 2;
static const int BishopCentreEndgame = 3;
static const int RookFileOpening = 3;
static const int QueenCentreEndgame = 4;
static const int KingCentreEndgame = 12;
static const int KingFileOpening = 10;
static const int KingRankOpening = 10;

/* Adjustments */
static const int PawnCenterOpening = 20;
static const int KnightTrapped = 100;
static const int BishopDiagonalOpening = 4;
static const int BishopBackRankOpening = 10;
static const int QueenBackRankOpening = 5;

/*static void print_psqbonus()
{
	for (int phase = Opening; phase <= EndGame; phase++)
		for (int piece = Pawn; piece <= King; piece++) {
			for (int r = Rank8; r >= Rank1; r--) {
				for (int f = FileA; f <= FileH; f++) {
					const int sq = square(r, f);
					printf("%d,", PSQTable[phase][piece][sq] - Material[phase][piece]);
				}
				printf("\n");
			}
			printf("\n");
		}
}*/

void init_psq()
{
	if (PSQInitialized) return;
	else PSQInitialized = true;
	
	for (int piece = Pawn; piece <= King; piece++)
		for (int sq = A1; sq <= H8; sq++) {
			const int r = rank(sq), f = file(sq);
			int op = 0, eg = 0;
			
			switch (piece) {
			case Pawn:
				op += Center[f] * PawnFileOpening;
				if (sq == D5 || sq == E5 || sq == D3 || sq == E3)
					op += PawnCenterOpening / 2;
				else if (sq == D4 || sq == E4)
					op += PawnCenterOpening;
				break;
			case Knight:
				op += (Center[r] + Center[f]) * KnightCentreOpening;
				eg += (Center[r] + Center[f]) * KnightCentreEndgame;
				op += KnightRank[r] * KnightRankOpening;
				op -= KnightTrapped * (sq == A8 || sq == H8);
				break;
			case Bishop:
				op += (Center[r] + Center[f]) * BishopCentreOpening;
				eg += (Center[r] + Center[f]) * BishopCentreEndgame;
				op -= BishopBackRankOpening * (r == Rank1);
				op += BishopDiagonalOpening * (7 == r + f || r == f);
				break;
			case Rook:
				op += Center[f] * RookFileOpening;
				break;
			case Queen:
				eg += (Center[r] + Center[f]) * QueenCentreEndgame;
				op -= QueenBackRankOpening * (r == Rank1);
				break;
			case King:
				eg += (Center[r] + Center[f]) * KingCentreEndgame;
				op += KingFile[f] * KingFileOpening + KingRank[r] * KingRankOpening;
				break;
			}
			
			PSQTable[Opening][piece][sq] = Material[Opening][piece] + op;
			PSQTable[EndGame][piece][sq] = Material[EndGame][piece] + eg;
		}
}

static int calc_psq(const uint64_t b[NB_PIECE], int phase, int color)
{
	assert(PSQInitialized && phase_ok(phase));
	int result = 0;

	for (int piece = Pawn; piece <= King; piece++) {
		uint64_t sqs = b[piece];
		while (sqs) {
			int sq = next_bit(&sqs);
			result += get_psqtab(color, phase, piece, sq);
		}
	}

	return result;
}

bool psq_ok(const Board *B)
{
	for (int phase = Opening; phase <= EndGame; phase++)
		for (int color = White; color <= Black; color++)
			if (B->st->psq[color][phase] != calc_psq(B->b[color], phase, color))
				return false;
	return true;
}
