/*
 * DoubleCheck, a UCI chess playing engine. Copyright (C) 2011 Lucas Braesch.
 *
 * DoubleCheck is free software: you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * DoubleCheck is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not,
 * see <http://www.gnu.org/licenses/>.
*/
#include "search.h"
#include "util.h"

bool Verbose = true;

#define Mate	30000
static int mate_in (int ply);
static int mated_in(int ply);
static bool is_mate_score(int score);

static const bool UseNull = true;
static const int NullDepth = 2;
static inline int NullReduction(int depth) {
	return max(4, depth/2);
}

static const bool UseRazoring = true;
static const int RazorDepth = 3;
static inline int RazorMargin(int depth) {
	assert(1 <= depth && depth <= RazorDepth);
	return 2*MP + (depth-1)*MP/4;
}

const bool UseEval = true;
static inline int EvalMargin(int depth) {
	assert(1 <= depth && depth <= RazorDepth);
	return depth * MN;
}

static const bool UseIID = true;
static const int IIDDepth[2] = {7, 4};
static const int IIDMargin = MP;

static const bool UseFutility = true;
static const bool UseLMR = true;

static int get_pv(Board *B, move_t *pv);
static void print_pv(Board *B, SearchInfo *si);

static void update_killers(const Board *B, SearchInfo *si);
static bool can_return_tt(bool is_pv, const hash_t *tte, int depth, int beta, int ply);
static int adjust_tt_score(int score, int ply);

#ifndef NDEBUG
static bool alpha_beta_ok(int alpha, int beta, bool is_pv)
{
	return -INF <= alpha && alpha < beta && beta <= +INF
	       && (is_pv || alpha+1 == beta);
}
#endif

static void init_node(SearchInfo *si);

bool abort_search;
uint64_t node_count;
static uint64_t node_limit;
static clock_t start_clock, time_clock;

move_t id_loop(Board *B, int max_time, uint64_t max_nodes, int max_depth)
{
	assert(B);
	move_t result = NoMove;

	start_clock = clock();
	time_clock = max_time * CLOCKS_PER_SEC / 1000;

	node_count = 0;
	node_limit = max_nodes;
	max_depth = max_depth ? max(min(max_depth, MAX_PLY-1), 1) : MAX_PLY-1;
	abort_search = false;

	SearchInfo si[MAX_PLY];
	memset(si, 0, sizeof(si));
	clear_history();
	tt_newsearch();

	static const int base_delta = 16;
	
	for (int depth = 1, alpha = -INF, beta = +INF; !abort_search && depth <= max_depth; depth++) {
		if (Verbose) my_send("info depth %d\n", depth);
		
		/* Aspiration loop */
		int score, delta = base_delta;
		do {
			score = search(B, alpha, beta, depth, 0, true, si);
			
			if (alpha < score && score < beta) {
				// score is within bounds
				if (depth >= 4 && !is_mate_score(score)) {
					// set aspiration window for the next depth (so aspiration starts at depth 5)
					alpha = score - delta;
					beta = score + delta;
				}
				// stop the aspiration loop
				break;
			} else {
				// resize window
				if (score <= alpha)
					alpha -= delta;
				else if (score >= beta)
					beta += delta;
				// incrase delta (following a geometric sequence with reason 1.5)
				delta += delta/2;
				// aspiration loop will continue, unless search is aborted
			}
		} while (!abort_search);
	
		if (!abort_search || depth==1) {
			my_send("info score cp %d depth %d nodes %"PRIu64" time %ld ", score, depth, node_count,
			       (clock()-start_clock)*1000/CLOCKS_PER_SEC);
			print_pv(B, si);
			result = si->best;
		}
	}

	return result;
}

int search(Board *B, int alpha, int beta, int depth, int ply, bool is_pv, SearchInfo *si)
{
	if (depth <= 0)
		return qsearch(B, alpha, beta, depth, ply, is_pv, si);

	assert(alpha_beta_ok(alpha, beta, is_pv));
	assert(depth > 0 && ply >= 0);
	const bool in_check = board_is_check(B);
	int best_idx = -INF, best_score = -INF, old_alpha = alpha;
	int current_eval;

	if ((board_is_draw(B) || ply >= MAX_PLY-2) && ply > 0)
		return 0;

	init_node(si);

	// mate distance pruning
	alpha = max(mated_in(ply), alpha);
	beta = min(mate_in(ply+1), beta);
	if (alpha >= beta)
		return alpha;

	// TT lookup
	uint64_t key = get_key(B);
	hash_t *tte = tt_find(key);
	if (tte) {
		if (tte->depth > 0)
			si->tt = tte->move;
		current_eval = tte->eval;
		if (ply > 0 && can_return_tt(is_pv, tte, depth, beta, ply)) {
			tt_refresh(tte);
			return adjust_tt_score(tte->score, ply);
		}
	} else
		current_eval = in_check ? -INF : calc_eval(B);

	// Razoring
	if (UseRazoring && depth <= RazorDepth
		&& !is_pv && !is_mate_score(beta) && !in_check)
	{
		const int threshold = beta - RazorMargin(depth);
		if (current_eval < threshold) {
			const int score = qsearch(B, threshold-1, threshold, 0, ply+1, is_pv, si+1);
			if (score < threshold)
				return score;
		}
	}

	// Eval pruning
	if (UseEval && depth <= RazorDepth
		&& !is_pv && !is_mate_score(beta) && !in_check
		&& current_eval - EvalMargin(depth) >= beta)
		return current_eval - EvalMargin(depth);
	
	// Null move pruning
	if (UseNull && depth >= NullDepth && !is_pv
		&& !in_check && !is_mate_score(beta)
		&& B->st->npm[B->turn]		// B->turn has pieces
		&& (depth <= NullReduction(depth) || current_eval >= beta))
	{
		const int new_depth = depth - NullReduction(depth);

		play(B, NoMove);
		int score = -search(B, -beta, -alpha, new_depth, ply+1, false, si+1);
		undo(B);

		if (score >= beta) {
			// null search fails high
			return score < mate_in(MAX_PLY)
				? score		// fail soft
				: beta;		// *but* do not return an unproven mate
		}
	}
	
	// Internal Iterative Deepening
	if (UseIID && depth >= IIDDepth[is_pv] && move_equal(si->tt, NoMove)
        && (is_pv || (!in_check && current_eval + IIDMargin >= beta)))
	{
		search(B, alpha, beta, is_pv ? depth-2 : depth/2, ply, is_pv, si);
		si->tt = si->best;
	}

	// generate all moves
	MoveSortList msl;
	MoveSort *ms;
	gen_msl(B, &msl, 0, si);
	
	while (alpha < beta && NULL != (ms = next_move(B, &msl)) && !abort_search) {
		int score;
		si->current = ms->m;
		const unsigned check = is_check(B, ms->m);
		if (ms->see == -INF) ms->see = see_move(B, ms->m);
		
		// root node, display move to search
		if (Verbose && !ply && node_count >= 0x10000) {
			char s[16];
			move_to_string(B, si->current, s, false);
			my_send("info currmove %s currmovenumber %d\n", s, msl.idx);
		}
		
		// search extensions
		int ext = (1 == msl.count)			// forced move
			|| (check == 1 && ms->see >= 0)	// normal check, SEE >= 0
			|| check == 2;					// discovered check, regardless of SEE
		int new_depth = depth - 1 + ext;

		// Futility Pruning
		if (UseFutility && !is_pv
			&& depth == 1 && !ext
			&& !check
			&& !in_check)
		{
			const int opt_score = current_eval + EP + optimistic_gain(B, si->current);

			if (opt_score <= alpha) {
				if (opt_score > best_score)
					best_score = opt_score;
				continue;
			}
		}

		// move properties
		const bool first = msl.idx == 1;
		const bool capture = is_cop(B, si->current);
		const bool killer = move_equal(si->current, si->killer[0]) || move_equal(si->current, si->killer[1]);
		const bool pthreat =
			is_passed_pawn_push(B, si->current)
			&& ms->see >= 0;
		const int hscore = capture ? 0 : get_history(B, si->current);

		// reduction decision
		const bool bad_capture = capture && ms->see < 0;
		const bool bad_quiet = !capture
			&& (hscore < 0 || (hscore == 0 && ms->see < 0));
		si->reduction = UseLMR
			&& !ext
			&& !first
			&& (bad_capture || bad_quiet)
			&& !killer
			&& !pthreat;
		if (msl.idx >= 4 && bad_quiet && !is_castling(B, si->current))
			si->reduction *= 2;

		play(B, si->current);

		if (is_pv && first) {
			// search full window
			score = -search(B, -beta, -alpha, new_depth, ply+1, is_pv, si+1);
		} else {
			// zero window search (reduced)
			score = -search(B, -alpha-1, -alpha, new_depth-si->reduction, ply+1, false, si+1);
			
			// doesn't fail low: re-search full window (reduced)
			if (score > alpha)
				score = -search(B, -beta, -alpha, new_depth-si->reduction, ply+1, is_pv, si+1); 

			// fails high: verify the beta cutoff with a non reduced search
			if (score >= beta && si->reduction)
				score = -search(B, -beta, -alpha, new_depth, ply+1, is_pv, si+1);
		}

		undo(B);

		assert(best_score <= alpha);
		// update search bounds
		if (score > best_score) {
			best_score = score;
			best_idx = msl.idx-1;
			si->best = si->current;
			if (score > alpha)
				alpha = score;
		}
	}

	// never store aborted search results anywhere
	if (abort_search)
		return best_score;

	if (!msl.count)		// mated or stalemate
		return in_check ? mated_in(ply) : 0;

	if (!ply) {						// stop id loop in trivial cases
		if (1 == msl.count)			// forced move
			abort_search = true;
		else if (best_score >= mate_in(depth) || best_score <= mated_in(depth))
			abort_search = true;	// mate score that can't be improved
	}

	// update TT
	unsigned type = best_score <= old_alpha ? ScoreUbound
		: (best_score >= beta ? ScoreLbound : ScoreExact);
	tt_write(key, best_score, current_eval, type, depth, si->best);

	// update killers and history (quiet moves only)
	if (best_idx >= 0 && !is_cop(B, si->best)) {
		update_killers(B, si);

		assert(move_equal(si->best, msl.ms[best_idx].m));
		// mark best move as good
		history_update(B, si->best, depth * depth);
		// mark all other moves searched as bad
		for (int i = 0; i < best_idx; i++) {
			move_t m = msl.ms[i].m;
			if (!is_cop(B, m))
				history_update(B, m, -depth * depth);
		}
	}

	return best_score;
}

int qsearch(Board *B, int alpha, int beta, int depth, int ply, bool is_pv, SearchInfo *si)
{
	assert(alpha_beta_ok(alpha, beta, is_pv));
	assert(depth <= 0 && ply >= 0);
	int best_score, old_alpha = alpha;
	bool in_check = board_is_check(B);

	init_node(si);

	// draw or MAX_PLY reached ?
	if (board_is_draw(B) || ply >= MAX_PLY-2)
		return 0;

	// mate distance pruning
	alpha = max(mated_in(ply), alpha);
	beta = min(mate_in(ply+1), beta);
	if (alpha >= beta)
		return alpha;

	// TT lookup
	uint64_t key = get_key(B);
	hash_t *tte = tt_find(key);
	int current_eval;
	if (tte) {
		if (can_return_tt(is_pv, tte, depth, beta, ply)) {
			tt_refresh(tte);
			return adjust_tt_score(tte->score, ply);
		}
		si->tt = tte->move;
		current_eval = tte->eval;
	} else
		current_eval = in_check ? -INF : calc_eval(B);

	// consider the current eval
	best_score = current_eval;
	if (best_score > alpha) {
		assert(!in_check);
		alpha = best_score;
		if (alpha >= beta)
			return alpha;
	}

	// generate captures only, or check escapes when in check
	MoveSortList msl;
	MoveSort *ms;
	gen_msl(B, &msl, depth < 0 ? 2 : 1, si);
	
	const int fut_base = current_eval + EP/4;

	while (alpha < beta && (ms = next_move(B, &msl)) != NULL && !abort_search) {
		int score;
		si->current = ms->m;
		const unsigned check = is_check(B, ms->m);
		if (ms->see == -INF) ms->see = see_move(B, ms->m);

		// Futility pruning
		if (UseFutility && !is_pv && !check && !in_check) {
			const int opt_score = fut_base + optimistic_gain(B, si->current);

			if (opt_score <= alpha) {
				if (opt_score > best_score)
					best_score = opt_score;
				continue;
			}

			if (fut_base <= alpha && depth < 0 && ms->see <= 0)
				continue;
		}

		// non-capture evasions that are candidate to be pruned (idea from StockFish)
		const bool evasion_prunable = in_check && best_score > mated_in(MAX_PLY)
			&& !is_cop(B, si->current);

		// prune negative SEE and prunable evasions
		if (!is_pv && ms->see < 0 && check < 2
			&& (!in_check || evasion_prunable))
			continue;

		play(B, si->current);
		score = -qsearch(B, -beta, -alpha, depth-1, ply+1, is_pv, si+1);
		undo(B);

		assert(best_score <= alpha);
		// beta cutoff
		if (score > best_score) {
			best_score = score;
			si->best = si->current;
			if (score > alpha)
				alpha = score;
		}
	}

	// never store aborted search results in the hash_table_t
	if (abort_search)
		return best_score;
	
	if (!msl.count) {
		// empty move loop: can be mated
		if (board_is_check(B))
			return mated_in(ply);
	} else {
		// update hash_table_t
		unsigned type = best_score <= old_alpha ? ScoreUbound
						: (best_score >= beta ? ScoreLbound : ScoreExact);
		tt_write(key, best_score, current_eval, type, depth < 0 ? -1 : 0, si->best);
	}

	return best_score;
}

static inline int mate_in (int ply) { return Mate - ply; }
static inline int mated_in(int ply) { return ply - Mate; }

static bool is_mate_score(int score)
{
	assert(abs(score) <= INF);
	return score <= mated_in(MAX_PLY) || score >= mate_in(MAX_PLY);
}

static inline int adjust_tt_score(int score, int ply)
/* mate scores from the hash_table_t need to be adjusted. For example, if we find a mate in 10 from the
 * hash_table_t at ply 5, then we effectively have a mate in 15 plies. */
{
	if (score >= mate_in(MAX_PLY))
		return score - ply;
	else if (score <= mated_in(MAX_PLY))
		return score + ply;
	else
		return score;
}

static inline bool can_return_tt(bool is_pv, const hash_t *tte, int depth, int beta, int ply)
/* PV nodes: return only exact scores
 * non PV nodes: return fail high/low scores. Mate scores are also trusted, regardless of the depth.
 * This idea is from StockFish, and although it's not totally sound, it seems to work.
 * */
{
	const bool depth_ok = tte->depth >= depth;

	if (is_pv)
		return depth_ok && hash_type(tte) == ScoreExact;
	else {
		const int tt_score = adjust_tt_score(tte->score, ply);
		return (depth_ok
		        ||	tt_score >= max(mate_in(MAX_PLY), beta)
		        ||	tt_score < min(mated_in(MAX_PLY), beta))
		       &&	((hash_type(tte) == ScoreLbound && tt_score >= beta)
		            ||(hash_type(tte) == ScoreUbound && tt_score < beta));
	}
}

static inline void update_killers(const Board *B, SearchInfo *si)
{
	assert(!is_cop(B, si->best));
	
	if (!move_equal(si->best, si->killer[0])) {
		si->killer[1] = si->killer[0];
		si->killer[0] = si->best;
	}
}

static int get_pv(Board *B, move_t *pv)
{
	int nb_moves = 0;

	for (int i = 0; i < (Verbose ? MAX_PLY : 1); i++) {
		const uint64_t key = get_key(B);
		const hash_t *tte = tt_find(key);

		if (!tte
			|| hash_type(tte) != ScoreExact
			|| !move_ok(B, tte->move)
			|| board_is_draw(B))
			break;

		play(B, tte->move);
		*pv++ = tte->move;
		nb_moves++;
	}

	for (int i = 0; i < nb_moves; i++)
		undo(B);

	return nb_moves;
}

static void print_pv(Board *B, SearchInfo *si)
{
	char s[16];
	move_t pv[MAX_PLY];
	int pv_length = get_pv(B, pv);

	my_send("pv");

	for (int i = 0; i < pv_length; i++) {
		move_to_string(B, pv[i], s, false);
		my_send(" %s", s);
	}
	my_send("\n");
}

static void init_node(SearchInfo *si)
// increment node count, and abort search if time is up
{
	si->best = si->current = si->tt = NoMove;
	si->reduction=0;

	if (++node_count % 0x200 == 0) {
		if (node_limit && node_count >= node_limit)
			abort_search = true;
		else if (time_clock && (clock() >= start_clock + time_clock))
			abort_search = true;
		else
			event();
	}
}
