/*
	DoubleCheck, a UCI chess playing engine. Copyright (C) 2011 Lucas Braesch.

	DoubleCheck is free software: you can redistribute it and/or modify it under the terms of the GNU
	General Public License as published by the Free Software Foundation, either version 3 of the
	License, or (at your option) any later version.

	DoubleCheck is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
	even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License along with this program.
	If not, see <http://www.gnu.org/licenses/>.
*/
#include "tt.h"
#include "util.h"
#include "move.h"

static hash_table_t TT = {0, NULL, 0};

void tt_alloc(uint64_t size)
/* allocates TT.count elements, where TT.count is the largest power of two such that the total size
 * sizeof(hash_t) * TT.count is less or equal to size bytes.
 * Note: This function can be re-called at will. It will internally do the right thing, either:
 * - nothing
 * - realloc and zero out any exceeding portion only (existing hash is preserved)
 * - alloc for the first time, and zero out the buffer
 */
{
	const uint64_t old_size = TT.count * sizeof(hash_t);
	
	// calculate the number of hash slots to allocate
	for (TT.count = 1; TT.count * sizeof(hash_t) <= size; TT.count *= 2);
	TT.count /= 2;
	
	// adjust size to a correct power of two
	size = sizeof(hash_t) * TT.count;
	
	// TT.buf already allocated to the correct size: nothing to do
	if (TT.count * sizeof(hash_t) == old_size)
		return;
	
	// use realloc to preserve the first min(size,old_size) bytes
	TT.buf = realloc(TT.buf, sizeof(hash_t) * TT.count);

	// when growing the buffer, the exceeding portion must be cleared
	if (size > old_size)
		memset(TT.buf + old_size, 0, size - old_size);
}

void tt_clear()
/* Note: It's OK to tt_clear() when the tt_alloc() has never been called, or when tt_free() has just
 * been called. The point is that, in all these cases, TT.buf may be NULL but TT.count is zero, so
 * the memset will do nothing, and will not attempt to deference a NULL pointer (C standard). */
{
	memset(TT.buf, 0, TT.count * sizeof(hash_t));
}

void tt_free()
/* Note: It's OK to free() a null pointer (C standard). So we can always use this even if the TT
 * hasn't been allocated (since TT.buf is initialized to NULL), and if tt_free() has already been
 * called. */
{
	free(TT.buf);
	TT.buf = NULL;
	TT.count = 0;
}

void tt_newsearch()
{
	TT.age++;
}

void tt_refresh(hash_t *tte)
{
	tte->age = TT.age;
}

void tt_write(uint64_t key, int score, int eval, int type, int depth, move_t move)
{
	assert(TT.count);
	hash_t *tte = &TT.buf[key & (TT.count - 1)];
	
	pack_t pack;
	pack.packed_key = key;
	pack.type = type;
	
	if (tte->pack.packed_key != pack.packed_key
		|| depth >= tte->depth
		|| tte->age != TT.age)
	{
		tte->pack = pack;
		tte->score = score;
		tte->eval = eval;
		tte->move = move;
		tte->depth = depth;
		tte->age = TT.age;
	}
}

hash_t *tt_find(uint64_t key)
{
	assert(TT.count);
	hash_t *tte = &TT.buf[key & (TT.count - 1)];
	
	pack_t pack;
	pack.packed_key = key;
	
	return tte->pack.packed_key == pack.packed_key ? tte : NULL;
}
