/*
 * DoubleCheck, a UCI chess playing engine. Copyright (C) 2011 Lucas Braesch.
 * 
 * DoubleCheck is free software: you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * DoubleCheck is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program. If not,
 * see <http://www.gnu.org/licenses/>.
*/
#include <string.h>
#include "search.h"
#include "uci_option.h"
#include "util.h"

static Board B;

static bool loop_step();
static void parse_go();
static void parse_option(const char *line);
static void parse_position(const char *line);

void main_loop()
{
	B.initialized = false;
	option_init();

	setvbuf(stdout, NULL, _IONBF, 0);

	while (loop_step());

	tt_free();
}

static bool loop_step()
// returns true, unless stdin pipe is closed or quit command issued
{
	char line[0x1000];
	if (!fgets(line, 0x1000, stdin))
		return false;	// pipe closed
	
	char *ptr = strchr(line,'\n');
	if (ptr) *ptr = '\0';
	const char *token = strtok(line, " ");
	if (!token) return false;

	if (!strcmp(token, "go")) {
		tt_alloc((uint64_t)option_get_int("Hash") << 20);
		parse_go();
	}
	else if (!strcmp(token, "uci")) {
		my_send("id name DoubleCheck 3.4\n");
		my_send("id author Lucas Braesch\n");
		option_list();
		my_send("uciok\n");
	}
	else if (!strcmp(token, "setoption"))
		parse_option(line + strlen("setoption "));
	else if (!strcmp(token, "ucinewgame"));	// nothing to do
	else if (!strcmp(token, "position"))
		parse_position(line + strlen("position "));
	else if (!strcmp(token, "isready"))
		my_send("readyok\n");
	else if (!strcmp(token, "quit"))
		return false;
	else if (!strcmp(token, "stop"))
		abort_search = true;
	else if (!strcmp(token, "eval")) {
		print_board(&B);
		char fen[MAX_FEN];
		write_fen(&B, fen);
		my_send("fen: %s\n", fen);
		print_eval(&B);
	} else if (!strcmp(token, "perft")) {
		// extract depth from the string
		const char *d = strtok(line + strlen("perft "), " ");
		int depth = atoi(d);
		
		clock_t start = clock();
		uint64_t count = perft(&B, depth, 0);
		my_send("time(ms) = %ld\n", (clock()-start)*1000/CLOCKS_PER_SEC);
		
		my_send("perft(%d) = %"PRIu64"\n", depth, count);
	} else
		my_send("loop_step(): unknown command \"%s\"\n", token);

	return true;
}

static void parse_option(const char *line)
{
	assert(line);
	const char *name = strstr(line, "name ");
	char *value = strstr(line, "value ");

	if (!name || !value || name >= value) return; // ignore buttons

	value[-1] = '\0'; // HACK: terminate the name string
	name += strlen("name ");
	value += strlen("value ");

	option_set(name, value);

	if (!strcmp(name, "Hash"))
		tt_alloc((uint64_t)option_get_int("Hash") << 20);
	else if (!strcmp(name, "Verbose"))
		Verbose = option_get_bool("Verbose");
}

static void parse_position(const char *line)
{
	assert(line);
	const char *fen = strstr(line, "fen ");
	const char *moves = strstr(line, "moves ");

	if (fen)
		load_fen(&B, fen + strlen("fen "));
	else	// assume startpos
		load_fen(&B, StartFen);

	if (moves) {
		const char *ptr = moves + strlen("moves ");
		char s[8];

		while (*ptr != '\0') {
			s[0] = *ptr++;
			s[1] = *ptr++;
			s[2] = *ptr++;
			s[3] = *ptr++;

			if (*ptr == '\0' || *ptr == ' ')
				s[4] = '\0';
			else {  	// promotion
				s[4] = *ptr++;
				s[5] = '\0';
			}

			move_t m = string_to_move(&B, s);
			play(&B, m);

			while (*ptr == ' ') ptr++;
		}
	}
}

static void parse_go()
{
	const char *token;
	unsigned wtime = 0, btime = 0, winc = 0, binc = 0;
	unsigned move_time = 0, max_depth = 0;
	uint64_t max_nodes = 0;

	#define ARG_ERROR(param) { my_send("parse_go(): missing argument for \"%s\"\n", param); return; }

	for (token = strtok(NULL," "); token; token = strtok(NULL," ")) {
		if (!strcmp(token, "wtime")) {
			if (!(token = strtok(NULL, " "))) ARG_ERROR("wtime");
			wtime = atoi(token);
		} else if (!strcmp(token, "btime")) {
			if (!(token = strtok(NULL, " "))) ARG_ERROR("btime");
			btime = atoi(token);
		} else if (!strcmp(token, "winc")) {
			if (!(token = strtok(NULL, " "))) ARG_ERROR("winc");
			winc = atoi(token);
		} else if (!strcmp(token, "binc")) {
			if (!(token = strtok(NULL, " "))) ARG_ERROR("binc");
			binc = atoi(token);
		} else if (!strcmp(token, "movetime")) {
			if (!(token = strtok(NULL, " "))) ARG_ERROR("movetime");
			move_time = atoi(token);
		} else if (!strcmp(token, "nodes")) {
			if (!(token = strtok(NULL, " "))) ARG_ERROR("nodes");
			max_nodes = atoll(token);
		} else if (!strcmp(token, "depth")) {
			if (!(token = strtok(NULL, " "))) ARG_ERROR("depth");
			max_depth = atoi(token);
		} else if (!strcmp(token, "infinite")) {
			// nothing to do: this is a non condition
		} else
			my_send("parse_go(): unknown parameter \"%s\"\n", token);
	}

	#undef ARG_ERROR

	const int blitz_time = B.turn ? btime : wtime;
	const int blitz_inc = B.turn ? binc : winc;
	const int max_time = blitz_time + blitz_inc > 0
		? min(blitz_time/40 + blitz_inc, blitz_time-25)
		: move_time;

	move_t best = id_loop(&B, max_time, max_nodes, max_depth);

	char s[8];
	move_to_string(&B, best, s, false);
	my_send("bestmove %s\n", s);
}

void event()
{
	while (!abort_search && input_available())
		loop_step();
}
