/*
 * DoubleCheck, a UCI chess playing engine. Copyright (C) 2011 Lucas Braesch.
 * 
 * DoubleCheck is free software: you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * DoubleCheck is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program. If not,
 * see <http://www.gnu.org/licenses/>.
 * 
 * Credits: 64-bit KISS RNG by Heinz Van Saanen
*/
#include "zobrist.h"

uint64_t zobrist[NB_COLOR][NB_PIECE][NB_SQUARE],
    zobrist_turn,
    zobrist_ep[NB_SQUARE],
    zobrist_castle[16];
bool ZobristInitialized = false;

static inline uint64_t rol(uint64_t x, unsigned k)
{ return (x << k) | (x >> (64 - k)); }

uint64_t rand64()
// 64-bit KISS RNG by Heinz Van Saanen
{
	static uint64_t	// seeds
		a = 0x5bc5cd8748be9fe8,
		b = 0x5164ed10aa17c81,
		c = 0x105730377a2a0e9c,
		d = 0xe6b137acae0d8b76;
	
	const uint64_t e = a - rol(b,  7);
	a = b ^ rol(c, 13);
	b = c + rol(d, 37);
	c = d + e;
	return d = e + a;
}

void init_zobrist()
{
	if (ZobristInitialized) return;
	else ZobristInitialized = true;

	for (unsigned c = White; c <= Black; c++)
		for (unsigned p = Pawn; p <= King; p++)
			for (unsigned sq = A1; sq <= H8; zobrist[c][p][sq++] = rand64());

	zobrist_turn = rand64();
	for (unsigned crights = 0; crights < 16; zobrist_castle[crights++] = rand64());
	for (unsigned sq = A1; sq <= H8; zobrist_ep[sq++] = rand64());
}

uint64_t calc_key(const Board *B)
{
	uint64_t key = 0ULL;

	for (unsigned color = White; color <= Black; color++)
		for (unsigned piece = Pawn; piece <= King; piece++) {
			uint64_t sqs = B->b[color][piece];
			while (sqs) {
				const unsigned sq = next_bit(&sqs);
				key ^= zobrist[color][piece][sq];
			}
		}

	return B->turn ? key ^ zobrist_turn : key;
}

uint64_t calc_kpkey(const Board *B)
{
	uint64_t kpkey = 0ULL;

	for (unsigned color = White; color <= Black; color++) {
		uint64_t sqs = B->b[color][Pawn];
		while (sqs) {
			const unsigned sq = next_bit(&sqs);
			kpkey ^= zobrist[color][Pawn][sq];
		}
		kpkey ^= zobrist[color][King][B->king_pos[color]];
	}
	
	return kpkey;
}